from typing import NewType

# Strings are taken from the users, but then tainted as this type for stricter type-checking:
# to prevent usage of some other strings (e.g. operator id) as the handlers ids.
# It is so much ubiquitous that it deserves its own module to avoid circular dependencies.
HandlerId = NewType('HandlerId', str)



import asyncio
import enum
import threading
from collections.abc import Awaitable, Generator
from typing import Generic, TypeVar

FlagReasonT = TypeVar('FlagReasonT', bound=enum.Flag)


class FlagSetter(Generic[FlagReasonT]):
    """
    A boolean flag indicating that the daemon should stop and exit.

    Every daemon gets a ``stopped`` kwarg, which is an event-like object.
    The stopped flag is raised in two cases:

    * The corresponding k8s object is deleted, so the daemon should stop.
    * The whole operator is stopping, so all the daemons should stop too.

    The stopped flag is a graceful way of a daemon termination.
    If the daemons do not react to their stoppers and continue running,
    their tasks are cancelled by raising a ``asyncio.CancelledError``.

    .. warning::
        In case of synchronous handlers, which are executed in the threads,
        this can lead to the OS resource leakage:
        there is no way to kill a thread in Python, so it will continue running
        forever or until failed (e.g. on an API call for an absent resource).
        The orphan threads will block the operator's process from exiting,
        thus affecting the speed of restarts.
    """

    def __init__(self) -> None:
        super().__init__()
        self.when: float | None = None
        self.reason: FlagReasonT | None = None
        self.sync_event = threading.Event()
        self.async_event = asyncio.Event()
        self.sync_waiter: SyncFlagWaiter[FlagReasonT] = SyncFlagWaiter(self)
        self.async_waiter: AsyncFlagWaiter[FlagReasonT] = AsyncFlagWaiter(self)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.is_set()}, reason={self.reason}>'

    def is_set(self, reason: FlagReasonT | None = None) -> bool:
        """
        Check if the daemon stopper is set: at all or for a specific reason.
        """
        matching_reason = reason is None or (self.reason is not None and reason in self.reason)
        return matching_reason and self.sync_event.is_set()

    def set(self, reason: FlagReasonT | None = None) -> None:
        reason = reason if reason is not None else self.reason  # to keep existing values
        self.when = self.when if self.when is not None else asyncio.get_running_loop().time()
        self.reason = reason if self.reason is None or reason is None else self.reason | reason
        self.sync_event.set()
        self.async_event.set()  # it is thread-safe: always called in operator's event loop.


class FlagWaiter(Generic[FlagReasonT]):
    """
    A minimalistic read-only checker for the daemons from the user side.

    This object is fed into the :kwarg:`stopped` kwarg for the handlers.

    The flag setter is hidden from the users, and is an internal class.
    The users should not be able to trigger the stopping activities.

    Usage:

    .. code-block:: python

        import kopf

        @kopf.daemon('kopfexamples')
        def handler(stopped, **kwargs):
            while not stopped:
                ...  # do something useful
                stopped.wait(60)
    """

    def __init__(self, setter: FlagSetter[FlagReasonT]) -> None:
        super().__init__()
        self._setter = setter

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.is_set()}, reason={self.reason}>'

    def __bool__(self) -> bool:
        return self._setter.is_set()

    def is_set(self) -> bool:
        return self._setter.is_set()

    @property
    def reason(self) -> FlagReasonT | None:
        return self._setter.reason

    # See the docstring for AsyncFlagPromise for explanation.
    def wait(self, timeout: float | None = None) -> "FlagWaiter[FlagReasonT]":
        # Presumably, `await stopped.wait(n).wait(m)` in async mode.
        raise NotImplementedError("Please report the use-case in the issue tracker if needed.")

    # See the docstring for AsyncFlagPromise for explanation.
    def __await__(self) -> Generator[None, None, "FlagWaiter[FlagReasonT]"]:
        # Presumably, `await stopped` in either sync or async mode.
        raise NotImplementedError("Daemon stoppers should not be awaited directly. "
                                  "Use `await stopped.wait()`.")


class SyncFlagWaiter(FlagWaiter[FlagReasonT], Generic[FlagReasonT]):
    def wait(self, timeout: float | None = None) -> "SyncFlagWaiter[FlagReasonT]":
        self._setter.sync_event.wait(timeout=timeout)
        return self


class AsyncFlagWaiter(FlagWaiter[FlagReasonT], Generic[FlagReasonT]):
    def wait(self, timeout: float | None = None) -> "AsyncFlagPromise[FlagReasonT]":
        # A new checker instance, which is awaitable and returns the original checker in the end.
        return AsyncFlagPromise(self, timeout=timeout)


class AsyncFlagPromise(FlagWaiter[FlagReasonT],
                       Awaitable[AsyncFlagWaiter[FlagReasonT]],
                       Generic[FlagReasonT]):
    """
    An artificial future-like promise for ``await stopped.wait(...)``.

    This is a low-level "clean hack" to simplify the publicly faced types.
    As a trade-off, the complexity moves to the implementation side.

    The only goal is to accept one and only one type for the ``stopped`` kwarg
    in ``@kopf.daemon()`` handlers (protocol :class:`callbacks.DaemonFn`)
    regardless of whether they are sync or async, but still usable for both.

    For this, the kwarg is announced with the base type :class:`DaemonStopped`,
    not as a union of sync or async, or any other double-typed approach.

    The challenge is to support ``stopped.wait()`` in both sync & async modes:

    * sync: ``stopped.wait(float) -> bool``.
    * async: ``await stopped.wait(float) -> bool``.

    The sync case is the primary case as there are no alternatives to it.
    The async case is secondary because such use is discouraged (but supported);
    it is recommended to rely on regular task cancellations in async daemons.

    To solve it, instead of returning a ``bool``, a ``bool``-evaluable object
    is returned --- the checker itself. This solves the primary (sync) use-case:
    just sleep for the requested duration and return bool-evaluable self.

    To follow the established signatures of the primary (sync) use-case,
    the secondary (async) use-case also returns an instance of the checker:
    but not "self"! Instead, it creates a new time-limited & awaitable checker
    for every call to ``stopped.wait(n)``, limiting its life by ``n`` seconds.

    Extra checker instances add some tiny memory overhead, but this is fine
    since the use-case is discouraged and there is a better native alternative.

    Then, going back through the class hierarchy, all classes are made awaitable
    so that this functionality becomes exposed via the declared base class.
    But all checkers except the time-limited one prohibit waiting for them.
    """

    def __init__(self, waiter: AsyncFlagWaiter[FlagReasonT], *, timeout: float | None) -> None:
        super().__init__(waiter._setter)
        self._timeout = timeout
        self._waiter = waiter

    def __await__(self) -> Generator[None, None, AsyncFlagWaiter[FlagReasonT]]:
        name = f"time-limited waiting for the daemon stopper {self._setter!r}"
        coro = asyncio.wait_for(self._setter.async_event.wait(), timeout=self._timeout)
        task = asyncio.create_task(coro, name=name)
        try:
            yield from task
        except asyncio.TimeoutError:
            pass  # the requested time limit is reached, exit regardless of the state
        return self._waiter  # the original checker! not the time-limited one!



import kopf


def _never1(*_, **__):
    return False


def _never2(*_, **__):
    return False


def _always1(*_, **__):
    return True


def _always2(*_, **__):
    return True


def test_notfn_when_true():
    combined = kopf.not_(_always1)
    result = combined()
    assert result is False


def test_notfn_when_false():
    combined = kopf.not_(_never1)
    result = combined()
    assert result is True


def test_allfn_when_all_are_true():
    combined = kopf.all_([_always1, _always2])
    result = combined()
    assert result is True


def test_allfn_when_one_is_false():
    combined = kopf.all_([_always1, _never1])
    result = combined()
    assert result is False


def test_allfn_when_all_are_false():
    combined = kopf.all_([_never1, _never2])
    result = combined()
    assert result is False


def test_allfn_when_no_functions():
    combined = kopf.all_([])
    result = combined()
    assert result is True


def test_anyfn_when_all_are_true():
    combined = kopf.any_([_always1, _always2])
    result = combined()
    assert result is True


def test_anyfn_when_one_is_false():
    combined = kopf.any_([_always1, _never1])
    result = combined()
    assert result is True


def test_anyfn_when_all_are_false():
    combined = kopf.any_([_never1, _never2])
    result = combined()
    assert result is False


def test_anyfn_when_no_functions():
    combined = kopf.any_([])
    result = combined()
    assert result is False


def test_nonefn_when_all_are_true():
    combined = kopf.none_([_always1, _always2])
    result = combined()
    assert result is False


def test_nonefn_when_one_is_false():
    combined = kopf.none_([_always1, _never1])
    result = combined()
    assert result is False


def test_nonefn_when_all_are_false():
    combined = kopf.none_([_never1, _never2])
    result = combined()
    assert result is True


def test_nonefn_when_no_functions():
    combined = kopf.none_([])
    result = combined()
    assert result is True



import certvalidator
import pytest

from kopf._kits.webhooks import WebhookServer


def test_missing_oscrypto(no_oscrypto):
    with pytest.raises(ImportError) as err:
        WebhookServer.build_certificate(['...'])
    assert "pip install certbuilder" in str(err.value)


def test_missing_certbuilder(no_certbuilder):
    with pytest.raises(ImportError) as err:
        WebhookServer.build_certificate(['...'])
    assert "pip install certbuilder" in str(err.value)


def test_certificate_generation():
    names = ['hostname1', 'hostname2', '1.2.3.4', '0:0:0:0:0:0:0:1']
    cert, pkey = WebhookServer.build_certificate(names)
    context = certvalidator.ValidationContext(extra_trust_roots=[cert])
    validator = certvalidator.CertificateValidator(cert, validation_context=context)
    path = validator.validate_tls('hostname1')
    assert len(path) == 1  # self-signed
    assert path.first.ca
    assert path.first.self_issued
    assert set(path.first.valid_domains) == {'hostname1', 'hostname2', '1.2.3.4', '::1'}
    assert set(path.first.valid_ips) == {'1.2.3.4', '::1'}


@pytest.mark.parametrize('hostnames, common_name', [
    (['hostname1', 'hostname2'], 'hostname1'),
    (['hostname2', 'hostname1'], 'hostname2'),
    (['1.2.3.4', 'hostname1'], 'hostname1'),
    (['1.2.3.4', '2.3.4.5'], '1.2.3.4'),
])
def test_common_name_selection(hostnames, common_name):
    cert, pkey = WebhookServer.build_certificate(hostnames)
    context = certvalidator.ValidationContext(extra_trust_roots=[cert])
    validator = certvalidator.CertificateValidator(cert, validation_context=context)
    path = validator.validate_tls(common_name)
    assert path.first.subject.native['common_name'] == common_name



import asyncio
import concurrent.futures
import threading
from typing import Any

from kopf._cogs.aiokits import aiotasks

Flag = aiotasks.Future | asyncio.Event | concurrent.futures.Future[Any] | threading.Event


async def wait_flag(
        flag: Flag | None,
) -> Any:
    """
    Wait for a flag to be raised.

    Non-asyncio primitives are generally not our worry,
    but we support them for convenience.
    """
    match flag:
        case None:
            return None
        case asyncio.Future():
            return await flag
        case asyncio.Event():
            return await flag.wait()
        case concurrent.futures.Future():
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(None, flag.result)
        case threading.Event():
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(None, flag.wait)
        case _:
            raise TypeError(f"Unsupported type of a flag: {flag!r}")


async def raise_flag(
        flag: Flag | None,
) -> None:
    """
    Raise a flag.

    Non-asyncio primitives are generally not our worry,
    but we support them for convenience.
    """
    match flag:
        case None:
            return None
        case asyncio.Future():
            flag.set_result(None)
        case asyncio.Event():
            flag.set()
        case concurrent.futures.Future():
            flag.set_result(None)
        case threading.Event():
            flag.set()
        case _:
            raise TypeError(f"Unsupported type of a flag: {flag!r}")


def check_flag(
        flag: Flag | None,
) -> bool | None:
    """
    Check if a flag is raised.
    """
    match flag:
        case None:
            return None
        case asyncio.Future():
            return flag.done()
        case asyncio.Event():
            return flag.is_set()
        case concurrent.futures.Future():
            return flag.done()
        case threading.Event():
            return flag.is_set()
        case _:
            raise TypeError(f"Unsupported type of a flag: {flag!r}")



import pytest

import kopf
from kopf._cogs.aiokits.aiovalues import Container
from kopf._cogs.clients.errors import APIConflictError, APIError, \
                                      APIForbiddenError, APIUnauthorizedError
from kopf._cogs.structs.references import MUTATING_WEBHOOK, VALIDATING_WEBHOOK
from kopf._core.engines.admission import configuration_manager
from kopf._core.intents.causes import WebhookType


@pytest.mark.parametrize('reason', set(WebhookType))
@pytest.mark.parametrize('selector', {VALIDATING_WEBHOOK, MUTATING_WEBHOOK})
async def test_nothing_happens_if_not_managed(
        mocker, settings, registry, insights, selector, reason, k8s_mocked):

    container = Container()
    mocker.patch.object(insights.ready_resources, 'wait')  # before the general Event.wait!
    mocker.patch.object(insights.backbone, 'wait_for')
    mocker.patch.object(container, 'as_changed')
    mocker.patch('asyncio.Event.wait')

    settings.admission.managed = None
    await configuration_manager(
        reason=reason,
        selector=selector,
        registry=registry,
        settings=settings,
        insights=insights,
        container=container,
    )

    assert not insights.ready_resources.wait.called
    assert not insights.backbone.wait_for.called
    assert not k8s_mocked.post.called
    assert not k8s_mocked.patch.called
    assert not container.as_changed.called


@pytest.mark.parametrize('reason', set(WebhookType))
@pytest.mark.parametrize('selector', {VALIDATING_WEBHOOK, MUTATING_WEBHOOK})
async def test_creation_is_attempted(
        mocker, settings, registry, insights, selector, resource, reason, k8s_mocked):

    container = Container()
    mocker.patch.object(container, 'as_changed', return_value=aiter([]))

    settings.admission.managed = 'xyz'
    await configuration_manager(
        reason=reason,
        selector=selector,
        registry=registry,
        settings=settings,
        insights=insights,
        container=container,
    )

    assert k8s_mocked.post.call_count == 1
    assert k8s_mocked.post.call_args_list[0][1]['url'].startswith('/apis/admissionregistration.k8s.io/')
    assert k8s_mocked.post.call_args_list[0][1]['payload']['metadata']['name'] == 'xyz'


@pytest.mark.parametrize('reason', set(WebhookType))
@pytest.mark.parametrize('selector', {VALIDATING_WEBHOOK, MUTATING_WEBHOOK})
async def test_creation_ignores_if_exists_already(
        mocker, settings, registry, insights, selector, resource, reason, k8s_mocked):

    container = Container()
    mocker.patch.object(container, 'as_changed', return_value=aiter([]))
    k8s_mocked.post.side_effect = APIConflictError({}, status=409)

    settings.admission.managed = 'xyz'
    await configuration_manager(
        reason=reason,
        selector=selector,
        registry=registry,
        settings=settings,
        insights=insights,
        container=container,
    )

    assert k8s_mocked.post.call_count == 1
    assert k8s_mocked.post.call_args_list[0][1]['url'].startswith('/apis/admissionregistration.k8s.io/')
    assert k8s_mocked.post.call_args_list[0][1]['payload']['metadata']['name'] == 'xyz'


@pytest.mark.parametrize('error', {APIError, APIForbiddenError, APIUnauthorizedError})
@pytest.mark.parametrize('reason', set(WebhookType))
@pytest.mark.parametrize('selector', {VALIDATING_WEBHOOK, MUTATING_WEBHOOK})
async def test_creation_escalates_on_errors(
        mocker, settings, registry, insights, selector, resource, reason, k8s_mocked, error):

    container = Container()
    mocker.patch.object(container, 'as_changed', return_value=aiter([]))
    k8s_mocked.post.side_effect = error({}, status=400)

    with pytest.raises(error):
        settings.admission.managed = 'xyz'
        await configuration_manager(
            reason=reason,
            selector=selector,
            registry=registry,
            settings=settings,
            insights=insights,
            container=container,
        )

    assert k8s_mocked.post.call_count == 1
    assert k8s_mocked.post.call_args_list[0][1]['url'].startswith('/apis/admissionregistration.k8s.io/')
    assert k8s_mocked.post.call_args_list[0][1]['payload']['metadata']['name'] == 'xyz'


@pytest.mark.parametrize('reason', set(WebhookType))
@pytest.mark.parametrize('selector', {VALIDATING_WEBHOOK, MUTATING_WEBHOOK})
async def test_patching_on_changes(
        mocker, settings, registry, insights, selector, resource, reason, k8s_mocked):

    @kopf.on.validate(*resource, registry=registry)
    def fn_v(**_): pass

    @kopf.on.mutate(*resource, registry=registry)
    def fn_m(**_): pass

    container = Container()
    mocker.patch.object(container, 'as_changed', return_value=aiter([
        {'url': 'https://hostname1/'},
        {'url': 'https://hostname2/'},
    ]))

    settings.admission.managed = 'xyz'
    await configuration_manager(
        reason=reason,
        selector=selector,
        registry=registry,
        settings=settings,
        insights=insights,
        container=container,
    )

    assert k8s_mocked.patch.call_count == 3
    assert k8s_mocked.patch.call_args_list[0][1]['url'].startswith('/apis/admissionregistration.k8s.io/')
    assert k8s_mocked.patch.call_args_list[0][1]['url'].endswith('/xyz')
    assert k8s_mocked.patch.call_args_list[1][1]['url'].startswith('/apis/admissionregistration.k8s.io/')
    assert k8s_mocked.patch.call_args_list[1][1]['url'].endswith('/xyz')
    assert k8s_mocked.patch.call_args_list[2][1]['url'].startswith('/apis/admissionregistration.k8s.io/')
    assert k8s_mocked.patch.call_args_list[2][1]['url'].endswith('/xyz')

    patch = k8s_mocked.patch.call_args_list[0][1]['payload']
    assert patch['webhooks']
    assert patch['webhooks'][0]['clientConfig']['url'].startswith('https://hostname1/')
    assert patch['webhooks'][0]['rules']
    assert patch['webhooks'][0]['rules'][0]['resources'] == ['kopfexamples']

    patch = k8s_mocked.patch.call_args_list[1][1]['payload']
    assert patch['webhooks']
    assert patch['webhooks'][0]['clientConfig']['url'].startswith('https://hostname2/')
    assert patch['webhooks'][0]['rules']
    assert patch['webhooks'][0]['rules'][0]['resources'] == ['kopfexamples']


@pytest.mark.parametrize('reason', set(WebhookType))
@pytest.mark.parametrize('selector', {VALIDATING_WEBHOOK, MUTATING_WEBHOOK})
async def test_patching_purges_non_permanent_webhooks(
        mocker, settings, registry, insights, selector, resource, reason, k8s_mocked):

    @kopf.on.validate(*resource, registry=registry, persistent=False)
    def fn_v(**_): pass

    @kopf.on.mutate(*resource, registry=registry, persistent=False)
    def fn_m(**_): pass

    container = Container()
    mocker.patch.object(container, 'as_changed', return_value=aiter([
        {'url': 'https://hostname/'},
    ]))

    settings.admission.managed = 'xyz'
    await configuration_manager(
        reason=reason,
        selector=selector,
        registry=registry,
        settings=settings,
        insights=insights,
        container=container,
    )

    assert k8s_mocked.patch.call_count == 2
    patch = k8s_mocked.patch.call_args_list[-1][1]['payload']
    assert not patch['webhooks']


@pytest.mark.parametrize('reason', set(WebhookType))
@pytest.mark.parametrize('selector', {VALIDATING_WEBHOOK, MUTATING_WEBHOOK})
async def test_patching_leaves_permanent_webhooks(
        mocker, settings, registry, insights, selector, resource, reason, k8s_mocked):

    @kopf.on.validate(*resource, registry=registry, persistent=True)
    def fn_v(**_): pass

    @kopf.on.mutate(*resource, registry=registry, persistent=True)
    def fn_m(**_): pass

    container = Container()
    mocker.patch.object(container, 'as_changed', return_value=aiter([
        {'url': 'https://hostname/'},
    ]))

    settings.admission.managed = 'xyz'
    await configuration_manager(
        reason=reason,
        selector=selector,
        registry=registry,
        settings=settings,
        insights=insights,
        container=container,
    )

    assert k8s_mocked.patch.call_count == 2
    patch = k8s_mocked.patch.call_args_list[-1][1]['payload']
    assert patch['webhooks'][0]['clientConfig']['url'].startswith('https://hostname/')
    assert patch['webhooks'][0]['rules']
    assert patch['webhooks'][0]['rules'][0]['resources'] == ['kopfexamples']


async def aiter(src):
    for item in src:
        yield item



from kopf._cogs.clients.api import get_default_namespace


async def test_default_namespace_when_unset(mocker, enforced_context):
    mocker.patch.object(enforced_context, 'default_namespace', None)
    ns = await get_default_namespace()
    assert ns is None


async def test_default_namespace_when_set(mocker, enforced_context):
    mocker.patch.object(enforced_context, 'default_namespace', 'xyz')
    ns = await get_default_namespace()
    assert ns == 'xyz'



import re

from kopf._cogs.structs.references import select_specific_namespaces


def test_empty_pattern_list():
    names = select_specific_namespaces([])
    assert not names


def test_included_empty_string():
    names = select_specific_namespaces([''])
    assert names == {''}


def test_included_exact_strings():
    names = select_specific_namespaces(['ns2', 'ns1'])
    assert names == {'ns1', 'ns2'}


def test_excluded_multipatterns():
    names = select_specific_namespaces(['ns1,ns2'])
    assert not names


def test_excluded_globs():
    names = select_specific_namespaces(['n*s', 'n?s'])
    assert not names


def test_excluded_regexps():
    names = select_specific_namespaces([re.compile(r'ns1')])
    assert not names



import abc
import copy
import json
from collections.abc import Collection, Iterable
from typing import Any, cast

from kopf._cogs.configs import conventions
from kopf._cogs.structs import bodies, dicts, patches


class DiffBaseStorage(conventions.StorageKeyMarkingConvention,
                      conventions.StorageStanzaCleaner,
                      metaclass=abc.ABCMeta):
    """
    Store the base essence for diff calculations, i.e. last handled state.

    The "essence" is a snapshot of meaningful fields, which must be tracked
    to identify the actual changes on the object (or absence of such).

    Used in the handling routines to check if there were significant changes
    (i.e. not the internal and system changes, like the uids, links, etc),
    and to get the exact per-field diffs for the specific handler functions.

    Conceptually similar to how ``kubectl apply`` stores the applied state
    on any object, and then uses that for the patch calculation:
    https://kubernetes.io/docs/concepts/overview/object-management-kubectl/declarative-config/
    """

    def __init__(self, ignored_fields: Iterable[dicts.FieldSpec] | None = None) -> None:
        super().__init__()
        self.ignored_fields = list(ignored_fields or [])  # materialize the iterable

    def build(
            self,
            *,
            body: bodies.Body,
            extra_fields: Iterable[dicts.FieldSpec] | None = None,
    ) -> bodies.BodyEssence:
        """
        Extract only the relevant fields for the state comparisons.

        The framework ignores all the system fields (mostly from metadata)
        and the status senza completely. Except for some well-known and useful
        metadata, such as labels and annotations (except for sure garbage).

        A special set of fields can be provided even if they are supposed
        to be removed. This is used, for example, for handlers which react
        to changes in the specific fields in the status stanza,
        while the rest of the status stanza is removed.

        It is generally not a good idea to override this method in custom
        stores, unless a different definition of an object's essence is needed.
        """

        # Always use a copy, so that future changes do not affect the extracted essence.
        essence = cast(dict[Any, Any], copy.deepcopy(dict(body)))

        # The top-level identifying fields never change, so there is not need to track them.
        if 'apiVersion' in essence:
            del essence['apiVersion']
        if 'kind' in essence:
            del essence['kind']

        # Purge the whole stenzas with system info (extra-fields are restored below).
        if 'metadata' in essence:
            del essence['metadata']
        if 'status' in essence:
            del essence['status']

        # We want some selected metadata to be tracked implicitly.
        dicts.cherrypick(src=body, dst=essence, fields=[
            'metadata.labels',
            'metadata.annotations',  # but not all of them! deleted below.
        ], picker=copy.deepcopy)

        # But we do not want all the annotations, only the potentially useful ones.
        # Also exclude the annotations of other Kopf-based operators' storages.
        annotations = essence.get('metadata', {}).get('annotations', {})
        ignored_prefixes = self._detect_marked_prefixes(annotations)
        for annotation in list(annotations):
            if any(annotation.startswith(f'{prefix}/') for prefix in ignored_prefixes):
                del annotations[annotation]
            elif annotation == 'kubectl.kubernetes.io/last-applied-configuration':
                del annotations[annotation]

        # Restore all explicitly whitelisted extra-fields from the original body.
        dicts.cherrypick(src=body, dst=essence, fields=extra_fields, picker=copy.deepcopy)

        self.remove_empty_stanzas(cast(bodies.BodyEssence, essence))

        # Remove ignored fields if specified
        for ignored_field in self.ignored_fields:
            try:
                dicts.remove(essence, ignored_field)
            except TypeError:
                # If the field does not support item deletion, just skip it.
                pass

        return cast(bodies.BodyEssence, essence)

    @abc.abstractmethod
    def fetch(
            self,
            *,
            body: bodies.Body,
    ) -> bodies.BodyEssence | None:
        raise NotImplementedError

    @abc.abstractmethod
    def store(
            self,
            *,
            body: bodies.Body,
            patch: patches.Patch,
            essence: bodies.BodyEssence,
    ) -> None:
        raise NotImplementedError


class AnnotationsDiffBaseStorage(conventions.StorageKeyFormingConvention, DiffBaseStorage):

    def __init__(
            self,
            *,
            prefix: str = 'kopf.zalando.org',
            key: str = 'last-handled-configuration',
            ignored_fields: Iterable[dicts.FieldSpec] | None = None,
            v1: bool = True,  # will be switched to False a few releases later
    ) -> None:
        super().__init__(prefix=prefix, v1=v1, ignored_fields=ignored_fields)
        self.key = key

    def build(
            self,
            *,
            body: bodies.Body,
            extra_fields: Iterable[dicts.FieldSpec] | None = None,
    ) -> bodies.BodyEssence:
        essence = super().build(body=body, extra_fields=extra_fields)
        self.remove_annotations(essence, set(self.make_keys(self.key, body=body)))
        self.remove_empty_stanzas(essence)
        return essence

    def fetch(
            self,
            *,
            body: bodies.Body,
    ) -> bodies.BodyEssence | None:
        for full_key in self.make_keys(self.key, body=body):
            encoded = body.metadata.annotations.get(full_key, None)
            decoded = json.loads(encoded) if encoded is not None else None
            if decoded is not None:
                return cast(bodies.BodyEssence, decoded)
        return None

    def store(
            self,
            *,
            body: bodies.Body,
            patch: patches.Patch,
            essence: bodies.BodyEssence,
    ) -> None:
        encoded: str = json.dumps(essence, separators=(',', ':'))  # NB: no spaces
        encoded += '\n'  # for better kubectl presentation without wrapping (same as kubectl's one)
        for full_key in self.make_keys(self.key, body=body):
            patch.metadata.annotations[full_key] = encoded
        self._store_marker(prefix=self.prefix, patch=patch, body=body)


class StatusDiffBaseStorage(DiffBaseStorage):

    def __init__(
            self,
            *,
            name: str = 'kopf',
            field: dicts.FieldSpec = 'status.{name}.last-handled-configuration',
            ignored_fields: Iterable[dicts.FieldSpec] | None = None,
    ) -> None:
        super().__init__(ignored_fields=ignored_fields)
        self._name = name
        real_field = field.format(name=self._name) if isinstance(field, str) else field
        self._field = dicts.parse_field(real_field)

    @property
    def field(self) -> dicts.FieldPath:
        return self._field

    @field.setter
    def field(self, field: dicts.FieldSpec) -> None:
        real_field = field.format(name=self._name) if isinstance(field, str) else field
        self._field = dicts.parse_field(real_field)

    def build(
            self,
            *,
            body: bodies.Body,
            extra_fields: Iterable[dicts.FieldSpec] | None = None,
    ) -> bodies.BodyEssence:
        essence = super().build(body=body, extra_fields=extra_fields)

        # Work around an issue with mypy not treating TypedDicts as MutableMappings.
        essence_dict = cast(dict[Any, Any], essence)
        dicts.remove(essence_dict, self.field)

        return essence

    def fetch(
            self,
            *,
            body: bodies.Body,
    ) -> bodies.BodyEssence | None:
        encoded: str | None = dicts.resolve(body, self.field, None)
        essence: bodies.BodyEssence | None = json.loads(encoded) if encoded is not None else None
        return essence

    def store(
            self,
            *,
            body: bodies.Body,
            patch: patches.Patch,
            essence: bodies.BodyEssence,
    ) -> None:
        # Store as a single string instead of full dict -- to avoid merges and unexpected data.
        encoded: str = json.dumps(essence, separators=(',', ':'))  # NB: no spaces
        dicts.ensure(patch, self.field, encoded)


class MultiDiffBaseStorage(DiffBaseStorage):

    def __init__(
            self,
            storages: Collection[DiffBaseStorage],
    ) -> None:
        super().__init__()
        self.storages = storages

    def build(
            self,
            *,
            body: bodies.Body,
            extra_fields: Iterable[dicts.FieldSpec] | None = None,
    ) -> bodies.BodyEssence:
        essence = super().build(body=body, extra_fields=extra_fields)
        for storage in self.storages:
            # Let the individual stores to also clean the essence from their own fields.
            # For this, assume the the previous essence _is_ the body (what's left of it).
            essence = storage.build(body=bodies.Body(essence), extra_fields=extra_fields)
        return essence

    def fetch(
            self,
            *,
            body: bodies.Body,
    ) -> bodies.BodyEssence | None:
        for storage in self.storages:
            content = storage.fetch(body=body)
            if content is not None:
                return content
        return None

    def store(
            self,
            *,
            body: bodies.Body,
            patch: patches.Patch,
            essence: bodies.BodyEssence,
    ) -> None:
        for storage in self.storages:
            storage.store(body=body, patch=patch, essence=essence)



import pytest

import kopf
from kopf._cogs.structs.references import Resource
from kopf._core.engines.admission import build_webhooks


@pytest.fixture()
def handlers(resource, registry):

    @kopf.on.validate(*resource, registry=registry)
    def validate_fn(**_):
        pass

    @kopf.on.mutate(*resource, registry=registry)
    def mutate_fn(**_):
        pass

    return registry._webhooks.get_all_handlers()


@pytest.mark.parametrize('id, field, exp_name', [
    ('id', None, 'id.sfx'),
    ('id.', None, 'id..sfx'),
    ('id-', None, 'id-.sfx'),
    ('id_', None, 'id-.sfx'),
    ('id!', None, 'id21.sfx'),
    ('id%', None, 'id25.sfx'),
    ('id/sub', None, 'id.sub.sfx'),
    ('id', 'fld1.fld2', 'id.fld1.fld2.sfx'),
])
@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
def test_name_is_normalised(registry, resource, decorator, id, field, exp_name):

    @decorator(*resource, id=id, field=field, registry=registry)
    def fn(**_):
        pass

    webhooks = build_webhooks(
        registry._webhooks.get_all_handlers(),
        resources=[resource],
        name_suffix='sfx',
        client_config={})

    assert len(webhooks) == 1
    assert webhooks[0]['name'] == exp_name


@pytest.mark.parametrize('id, field, exp_url', [
    ('id', None, 'https://hostname/p1/p2/id'),
    ('id.', None, 'https://hostname/p1/p2/id.'),
    ('id-', None, 'https://hostname/p1/p2/id-'),
    ('id_', None, 'https://hostname/p1/p2/id_'),
    ('id!', None, 'https://hostname/p1/p2/id%21'),
    ('id%', None, 'https://hostname/p1/p2/id%25'),
    ('id/sub', None, 'https://hostname/p1/p2/id/sub'),
    ('id', 'fld1.fld2', 'https://hostname/p1/p2/id/fld1.fld2'),
])
@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
def test_url_is_suffixed(registry, resource, decorator, id, field, exp_url):

    @decorator(*resource, id=id, field=field, registry=registry)
    def fn(**_):
        pass

    webhooks = build_webhooks(
        registry._webhooks.get_all_handlers(),
        resources=[resource],
        name_suffix='sfx',
        client_config={'url': 'https://hostname/p1/p2'})

    assert len(webhooks) == 1
    assert webhooks[0]['clientConfig']['url'] == exp_url


@pytest.mark.parametrize('id, field, exp_path', [
    ('id', None, 'p1/p2/id'),
    ('id.', None, 'p1/p2/id.'),
    ('id-', None, 'p1/p2/id-'),
    ('id_', None, 'p1/p2/id_'),
    ('id!', None, 'p1/p2/id%21'),
    ('id%', None, 'p1/p2/id%25'),
    ('id/sub', None, 'p1/p2/id/sub'),
    ('id', 'fld1.fld2', 'p1/p2/id/fld1.fld2'),
])
@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
def test_path_is_suffixed(registry, resource, decorator, id, field, exp_path):

    @decorator(*resource, id=id, field=field, registry=registry)
    def fn(**_):
        pass

    webhooks = build_webhooks(
        registry._webhooks.get_all_handlers(),
        resources=[resource],
        name_suffix='sfx',
        client_config={'service': {'path': 'p1/p2'}})

    assert len(webhooks) == 1
    assert webhooks[0]['clientConfig']['service']['path'] == exp_path


@pytest.mark.parametrize('opts, key, val', [
    (dict(side_effects=False), 'sideEffects', 'None'),
    (dict(side_effects=True), 'sideEffects', 'NoneOnDryRun'),
    (dict(ignore_failures=False), 'failurePolicy', 'Fail'),
    (dict(ignore_failures=True), 'failurePolicy', 'Ignore'),
])
@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
def test_flat_options_are_mapped(registry, resource, decorator, opts, key, val):

    @decorator(*resource, registry=registry, **opts)
    def fn(**_):
        pass

    webhooks = build_webhooks(
        registry._webhooks.get_all_handlers(),
        resources=[resource],
        name_suffix='sfx',
        client_config={})

    assert len(webhooks) == 1
    assert webhooks[0][key] == val
    assert webhooks[0]['matchPolicy'] == 'Equivalent'
    assert webhooks[0]['timeoutSeconds'] == 30
    assert webhooks[0]['admissionReviewVersions'] == ['v1', 'v1beta1']


# Mind the different supported collection types for operations, all converted to JSON lists.
@pytest.mark.parametrize('opts, key, val', [
    (dict(), 'operations', ['*']),
    (dict(operations={'CREATE'}), 'operations', ['CREATE']),
    (dict(operations={'UPDATE'}), 'operations', ['UPDATE']),
    (dict(operations={'DELETE'}), 'operations', ['DELETE']),
    (dict(operations=['CREATE','UPDATE']), 'operations', ['CREATE','UPDATE']),
    (dict(operations=['CREATE','DELETE']), 'operations', ['CREATE','DELETE']),
    (dict(operations=['UPDATE','DELETE']), 'operations', ['UPDATE','DELETE']),
])
@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
def test_rule_options_are_mapped(registry, resource, decorator, opts, key, val):

    @decorator(*resource, registry=registry, **opts)
    def fn(**_):
        pass

    webhooks = build_webhooks(
        registry._webhooks.get_all_handlers(),
        resources=[resource],
        name_suffix='sfx',
        client_config={})

    assert len(webhooks) == 1
    assert len(webhooks[0]['rules']) == 1
    assert webhooks[0]['rules'][0][key] == val
    assert webhooks[0]['rules'][0]['scope'] == '*'
    assert webhooks[0]['rules'][0]['apiGroups'] == [resource.group]
    assert webhooks[0]['rules'][0]['apiVersions'] == [resource.version]
    assert webhooks[0]['rules'][0]['resources'] == [resource.plural]


@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
def test_subresource_specific(registry, resource, decorator):

    @decorator(*resource, registry=registry, subresource='xyz')
    def fn(**_):
        pass

    webhooks = build_webhooks(
        registry._webhooks.get_all_handlers(),
        resources=[resource],
        name_suffix='sfx',
        client_config={})

    assert len(webhooks) == 1
    assert len(webhooks[0]['rules']) == 1
    assert webhooks[0]['rules'][0]['resources'] == [f'{resource.plural}/xyz']


@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
def test_subresource_catchall(registry, resource, decorator):

    @decorator(*resource, registry=registry, subresource='*')
    def fn(**_):
        pass

    webhooks = build_webhooks(
        registry._webhooks.get_all_handlers(),
        resources=[resource],
        name_suffix='sfx',
        client_config={})

    assert len(webhooks) == 1
    assert len(webhooks[0]['rules']) == 1
    assert webhooks[0]['rules'][0]['resources'] == [f'{resource.plural}/*']


@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
def test_subresource_absent(registry, resource, decorator):

    @decorator(*resource, registry=registry)
    def fn(**_):
        pass

    webhooks = build_webhooks(
        registry._webhooks.get_all_handlers(),
        resources=[resource],
        name_suffix='sfx',
        client_config={})

    assert len(webhooks) == 1
    assert len(webhooks[0]['rules']) == 1
    assert webhooks[0]['rules'][0]['resources'] == [f'{resource.plural}']


@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
def test_multiple_handlers(registry, resource, decorator):

    @decorator(*resource, registry=registry)
    def fn1(**_):
        pass

    @decorator(*resource, registry=registry)
    def fn2(**_):
        pass

    webhooks = build_webhooks(
        registry._webhooks.get_all_handlers(),
        resources=[resource],
        name_suffix='sfx',
        client_config={})

    assert len(webhooks) == 2
    assert len(webhooks[0]['rules']) == 1
    assert len(webhooks[1]['rules']) == 1


@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
def test_irrelevant_resources_are_ignored(registry, resource, decorator):

    @decorator(*resource, registry=registry)
    def fn(**_):
        pass

    irrelevant_resource = Resource('grp', 'vers', 'plural')
    webhooks = build_webhooks(
        registry._webhooks.get_all_handlers(),
        resources=[irrelevant_resource],
        name_suffix='sfx',
        client_config={})

    assert len(webhooks) == 1
    assert len(webhooks[0]['rules']) == 0


@pytest.mark.parametrize('label_value, exp_expr', [
    (kopf.PRESENT, {'key': 'lbl', 'operator': 'Exists'}),
    (kopf.ABSENT, {'key': 'lbl', 'operator': 'DoesNotExist'}),
    ('val', {'key': 'lbl', 'operator': 'In', 'values': ['val']}),
])
@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
def test_labels_specific_filter(registry, resource, decorator, label_value, exp_expr):

    @decorator(*resource, registry=registry, labels={'lbl': label_value})
    def fn(**_):
        pass

    irrelevant_resource = Resource('grp', 'vers', 'plural')
    webhooks = build_webhooks(
        registry._webhooks.get_all_handlers(),
        resources=[irrelevant_resource],
        name_suffix='sfx',
        client_config={})

    assert len(webhooks) == 1
    assert webhooks[0]['objectSelector'] == {'matchExpressions': [exp_expr]}


@pytest.mark.parametrize('decorator', [kopf.on.validate, kopf.on.mutate])
def test_labels_callable_filter(registry, resource, decorator):

    @decorator(*resource, registry=registry, labels={'lbl': lambda *_, **__: None})
    def fn(**_):
        pass

    irrelevant_resource = Resource('grp', 'vers', 'plural')
    webhooks = build_webhooks(
        registry._webhooks.get_all_handlers(),
        resources=[irrelevant_resource],
        name_suffix='sfx',
        client_config={})

    assert len(webhooks) == 1
    assert webhooks[0]['objectSelector'] is None



from unittest.mock import Mock

from kopf._cogs.structs.bodies import Body
from kopf._cogs.structs.ephemera import Memo
from kopf._core.reactor.inventory import ResourceMemories, ResourceMemory

BODY: Body = {
    'metadata': {
        'uid': 'uid1',
    }
}


async def test_creation_with_defaults():
    ResourceMemory()


async def test_recalling_creates_when_absent():
    memories = ResourceMemories()
    memory = await memories.recall(BODY)
    assert isinstance(memory, ResourceMemory)


async def test_recalling_reuses_when_present():
    memories = ResourceMemories()
    memory1 = await memories.recall(BODY)
    memory2 = await memories.recall(BODY)
    assert memory1 is memory2


async def test_forgetting_deletes_when_present():
    memories = ResourceMemories()
    memory1 = await memories.recall(BODY)
    await memories.forget(BODY)

    # Check by recalling -- it should be a new one.
    memory2 = await memories.recall(BODY)
    assert memory1 is not memory2


async def test_forgetting_ignores_when_absent():
    memories = ResourceMemories()
    await memories.forget(BODY)


async def test_memo_is_autocreated():
    memories = ResourceMemories()
    memory = await memories.recall(BODY)
    assert isinstance(memory.memo, Memo)


async def test_memo_is_shallow_copied():

    class MyMemo(Memo):
        def __copy__(self):
            mock()
            return MyMemo()

    mock = Mock()
    memobase = MyMemo()
    memories = ResourceMemories()
    memory = await memories.recall(BODY, memobase=memobase)
    assert mock.call_count == 1
    assert memory.memo is not memobase



import pytest

from kopf._cogs.structs.dicts import remove


def test_existing_key():
    d = {'abc': {'def': {'hij': 'val', 'hello': 'world'}}}
    remove(d, ['abc', 'def', 'hij'])
    assert d == {'abc': {'def': {'hello': 'world'}}}


def test_unexisting_key_in_existing_dict():
    d = {'abc': {'def': {'hello': 'world'}}}
    remove(d, ['abc', 'def', 'hij'])
    assert d == {'abc': {'def': {'hello': 'world'}}}


def test_unexisting_key_in_unexisting_dict():
    d = {}
    remove(d, ['abc', 'def', 'hij'])
    assert d == {}


def test_parent_cascaded_deletion_up_to_the_root():
    d = {'abc': {'def': {'hij': 'val'}}}
    remove(d, ['abc', 'def', 'hij'])
    assert d == {}


def test_parent_cascaded_deletion_up_to_a_middle():
    d = {'abc': {'def': {'hij': 'val'}, 'hello': 'world'}}
    remove(d, ['abc', 'def', 'hij'])
    assert d == {'abc': {'hello': 'world'}}


def test_nonmapping_key():
    d = {'key': 'val'}
    with pytest.raises(TypeError):
        remove(d, ['key', 'sub'])


def test_empty_path():
    d = {}
    with pytest.raises(ValueError) as e:
        remove(d, [])
    assert "Removing a root of a dict is impossible" in str(e.value)



import kopf
import pykube
import yaml


@kopf.on.create('kopfexamples')
def create_fn(spec, **kwargs):

    # Render the pod yaml with some spec fields used in the template.
    doc = yaml.safe_load(f"""
        apiVersion: v1
        kind: Pod
        spec:
          containers:
          - name: the-only-one
            image: busybox
            command: ["sh", "-x", "-c"]
            args:
            - |
              echo "FIELD=$FIELD"
              sleep {spec.get('duration', 0)}
            env:
            - name: FIELD
              value: {spec.get('field', 'default-value')}
    """)

    # Make it our child: assign the namespace, name, labels, owner references, etc.
    kopf.adopt(doc)

    # Actually create an object by requesting the Kubernetes API.
    api = pykube.HTTPClient(pykube.KubeConfig.from_env())
    pod = pykube.Pod(api, doc)
    pod.create()
    api.session.close()

    # Update the parent's status.
    return {'children': [pod.metadata['uid']]}



import asyncio
from collections.abc import Collection, Mapping

from kopf._cogs.clients import api, errors
from kopf._cogs.configs import configuration
from kopf._cogs.helpers import typedefs
from kopf._cogs.structs import references


async def read_version(
        *,
        settings: configuration.OperatorSettings,
        logger: typedefs.Logger,
) -> Mapping[str, str]:
    rsp: Mapping[str, str] = await api.get('/version', settings=settings, logger=logger)
    return rsp


async def scan_resources(
        *,
        settings: configuration.OperatorSettings,
        logger: typedefs.Logger,
        groups: Collection[str] | None = None,
) -> Collection[references.Resource]:
    coros = {
        _read_old_api(groups=groups, settings=settings, logger=logger),
        _read_new_apis(groups=groups, settings=settings, logger=logger),
    }
    resources: set[references.Resource] = set()
    for coro in asyncio.as_completed(coros):
        resources.update(await coro)
    return resources


async def _read_old_api(
        *,
        settings: configuration.OperatorSettings,
        logger: typedefs.Logger,
        groups: Collection[str] | None,
) -> Collection[references.Resource]:
    resources: set[references.Resource] = set()
    if groups is None or '' in groups:
        rsp = await api.get('/api', settings=settings, logger=logger)
        coros = {
            _read_version(
                url=f'/api/{version_name}',
                group='',
                version=version_name,
                preferred=True,
                settings=settings,
                logger=logger,
            )
            for version_name in rsp['versions']
        }
        for coro in asyncio.as_completed(coros):
            resources.update(await coro)
    return resources


async def _read_new_apis(
        *,
        settings: configuration.OperatorSettings,
        logger: typedefs.Logger,
        groups: Collection[str] | None,
) -> Collection[references.Resource]:
    resources: set[references.Resource] = set()
    if groups is None or set(groups or {}) - {''}:
        rsp = await api.get('/apis', settings=settings, logger=logger)
        items = [d for d in rsp['groups'] if groups is None or d['name'] in groups]
        coros = {
            _read_version(
                url=f'/apis/{group_dat["name"]}/{version["version"]}',
                group=group_dat['name'],
                version=version['version'],
                preferred=version['version'] == group_dat['preferredVersion']['version'],
                settings=settings,
                logger=logger,
            )
            for group_dat in items
            for version in group_dat['versions']
        }
        for coro in asyncio.as_completed(coros):
            resources.update(await coro)
    return resources


async def _read_version(
        *,
        url: str,
        group: str,
        version: str,
        preferred: bool,
        settings: configuration.OperatorSettings,
        logger: typedefs.Logger,
) -> Collection[references.Resource]:
    try:
        rsp = await api.get(url, settings=settings, logger=logger)
    except errors.APINotFoundError:
        # This happens when the last and the only resource of a group/version
        # has been deleted, the whole group/version is gone, and we rescan it.
        return set()
    else:
        # Note: builtins' singulars are empty strings in K3s (reasons unknown):
        # fall back to the lowercased kind so that the selectors could match.
        return {
            references.Resource(
                group=group,
                version=version,
                kind=resource['kind'],
                plural=resource['name'],
                singular=resource['singularName'] or resource['kind'].lower(),
                shortcuts=frozenset(resource.get('shortNames', [])),
                categories=frozenset(resource.get('categories', [])),
                subresources=frozenset(
                    subresource['name'].split('/', 1)[-1]
                    for subresource in rsp.get('resources', [])
                    if subresource['name'].startswith(f'{resource["name"]}/')
                ),
                namespaced=resource['namespaced'],
                preferred=preferred,
                verbs=frozenset(resource.get('verbs') or []),
            )
            for resource in rsp.get('resources', [])
            if '/' not in resource['name']
        }



import pathlib

import kopf

ROOT = (pathlib.Path.cwd() / pathlib.Path(__file__)).parent.parent.parent


@kopf.on.startup()
def config(settings: kopf.OperatorSettings, **_):

    # Plain and simple local endpoint with an auto-generated certificate:
    settings.admission.server = kopf.WebhookServer()

    # Plain and simple local endpoint with with provided certificate (e.g. openssl):
    settings.admission.server = kopf.WebhookServer(certfile=ROOT/'cert.pem', pkeyfile=ROOT/'key.pem', port=1234)

    # K3d/K3s-specific server that supports accessing from inside of a VM (a generated certificate):
    settings.admission.server = kopf.WebhookK3dServer(cadump=ROOT/'ca.pem')

    # K3d/K3s-specific server that supports accessing from inside of a VM (a provided certificate):
    settings.admission.server = kopf.WebhookK3dServer(certfile=ROOT/'k3d-cert.pem', pkeyfile=ROOT/'k3d-key.pem', port=1234)

    # Minikube-specific server that supports accessing from inside of a VM (a generated certificate):
    settings.admission.server = kopf.WebhookMinikubeServer(port=1234, cadump=ROOT/'ca.pem')

    # DockerDesktop-specific server that supports accessing from the host:
    settings.admission.server = kopf.WebhookDockerDesktopServer(port=1234)

    # Tunneling Kubernetes->ngrok->local server (anonymous, auto-loaded binary):
    settings.admission.server = kopf.WebhookNgrokTunnel(path='/xyz', port=1234)

    # Tunneling Kubernetes->ngrok->local server (registered users, pre-existing binary):
    settings.admission.server = kopf.WebhookNgrokTunnel(binary="/usr/local/bin/ngrok", token='...', )

    # Tunneling Kubernetes->ngrok->local server (registered users, pre-existing binary, specific region):
    settings.admission.server = kopf.WebhookNgrokTunnel(binary="/usr/local/bin/ngrok", region='eu')

    # Auto-detect the best server (K3d/Minikube/simple) strictly locally:
    settings.admission.server = kopf.WebhookAutoServer()

    # Auto-detect the best server (K3d/Minikube/simple) with external tunneling as a fallback:
    settings.admission.server = kopf.WebhookAutoTunnel()

    # The final configuration for CI/CD (overrides previous values):
    settings.admission.server = kopf.WebhookAutoServer()
    settings.admission.managed = 'auto.kopf.dev'


@kopf.on.validate('kex')
def authhook(headers: kopf.Headers, sslpeer: kopf.SSLPeer, warnings: list[str], **_):
    user_agent = headers.get('User-Agent', '(unidentified)')
    warnings.append(f"Accessing as user-agent: {user_agent}")
    if not sslpeer.get('subject'):
        warnings.append("SSL peer is not identified.")
    else:
        common_names = [val for key, val in sslpeer['subject'][0] if key == 'commonName']
        if common_names:
            warnings.append(f"SSL peer is {common_names[0]}.")
        else:
            warnings.append("SSL peer's common name is absent.")


@kopf.on.validate('kex')
def validate1(spec, dryrun, **_):
    if not dryrun and spec.get('field') == 'wrong':
        raise kopf.AdmissionError("Meh! I don't like it. Change the field.")


@kopf.on.validate('kex', field='spec.field', value='not-allowed')
def validate2(**_):
    raise kopf.AdmissionError("I'm too lazy anyway. Go away!", code=555)


@kopf.on.validate('kex', subresource='*')
def validate_subresources(spec, subresource, status, warnings: list[str], **_):
    if subresource == 'status' and status.get('field') != spec.get('field'):
        raise kopf.AdmissionError("status.field MUST be equal to spec.field!")
    elif subresource is None and status.get('field') != spec.get('field'):
        warnings.append("Also update status.field to match spec.field: "
                        f"{spec.get('field')!r} != {status.get('field')!r}")


@kopf.on.mutate('kex', labels={'somelabel': 'somevalue'})
def mutate1(patch: kopf.Patch, **_):
    patch.spec['injected'] = 123


# Marks for the e2e tests (see tests/e2e/test_examples.py):
# We do not care: pods can have 6-10 updates here.
E2E_SUCCESS_COUNTS: dict[str, int] = {}



import aiohttp.web
import pytest

from kopf._cogs.clients.creating import create_obj
from kopf._cogs.clients.errors import APIError


async def test_simple_body_with_arguments(
        resp_mocker, aresponses, hostname, settings, resource, namespace, logger, caplog):

    post_mock = resp_mocker(return_value=aiohttp.web.json_response({}))
    aresponses.add(hostname, resource.get_url(namespace=namespace), 'post', post_mock)

    body = {'x': 'y'}
    await create_obj(
        logger=logger,
        settings=settings,
        resource=resource,
        namespace=namespace,
        name='name1',
        body=body,
    )

    assert post_mock.called
    assert post_mock.call_count == 1

    data = post_mock.call_args_list[0][0][0].data  # [callidx][args/kwargs][argidx]
    if resource.namespaced:
        assert data == {'x': 'y', 'metadata': {'name': 'name1', 'namespace': 'ns'}}
    else:
        assert data == {'x': 'y', 'metadata': {'name': 'name1'}}


async def test_full_body_with_identifiers(
        resp_mocker, aresponses, hostname, settings, resource, namespace, caplog, logger):

    post_mock = resp_mocker(return_value=aiohttp.web.json_response({}))
    aresponses.add(hostname, resource.get_url(namespace=namespace), 'post', post_mock)

    body = {'x': 'y', 'metadata': {'name': 'name1', 'namespace': namespace}}
    await create_obj(
        logger=logger,
        settings=settings,
        resource=resource,
        body=body,
    )

    assert post_mock.called
    assert post_mock.call_count == 1

    data = post_mock.call_args_list[0][0][0].data  # [callidx][args/kwargs][argidx]
    assert data == {'x': 'y', 'metadata': {'name': 'name1', 'namespace': namespace}}


# Note: 401 is wrapped into a LoginError and is tested elsewhere.
@pytest.mark.parametrize('status', [400, 403, 404, 409, 500, 666])
async def test_raises_api_errors(
        resp_mocker, aresponses, hostname, settings, status, resource, namespace, logger,
        cluster_resource, namespaced_resource):

    post_mock = resp_mocker(return_value=aresponses.Response(status=status, reason='oops'))
    cluster_url = cluster_resource.get_url(namespace=None)
    namespaced_url = namespaced_resource.get_url(namespace='ns')
    aresponses.add(hostname, cluster_url, 'post', post_mock)
    aresponses.add(hostname, namespaced_url, 'post', post_mock)

    body = {'x': 'y'}
    with pytest.raises(APIError) as e:
        await create_obj(
            logger=logger,
            settings=settings,
            resource=resource,
            namespace=namespace,
            name='name1',
            body=body,
        )
    assert e.value.status == status



import aiohttp.web
import freezegun
import pytest

from kopf._core.engines.peering import Peer, clean, touch


@pytest.mark.parametrize('lastseen', [
    pytest.param('2020-01-01T00:00:00', id='when-dead'),
    pytest.param('2020-12-31T23:59:59', id='when-alive'),
])
@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_cleaning_peers_purges_them(
        hostname, aresponses, resp_mocker, settings, lastseen,
        peering_resource, peering_namespace):

    settings.peering.name = 'name0'
    patch_mock = resp_mocker(return_value=aiohttp.web.json_response({}))
    url = peering_resource.get_url(name='name0', namespace=peering_namespace)
    aresponses.add(hostname, url, 'patch', patch_mock)

    peer = Peer(identity='id1', lastseen=lastseen)
    await clean(peers=[peer], resource=peering_resource, settings=settings,
                namespace=peering_namespace)

    assert patch_mock.called
    patch = await patch_mock.call_args_list[0][0][0].json()
    assert set(patch['status']) == {'id1'}
    assert patch['status']['id1'] is None


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_touching_a_peer_stores_it(
        hostname, aresponses, resp_mocker, settings,
        peering_resource, peering_namespace):

    settings.peering.name = 'name0'
    patch_mock = resp_mocker(return_value=aiohttp.web.json_response({}))
    url = peering_resource.get_url(name='name0', namespace=peering_namespace)
    aresponses.add(hostname, url, 'patch', patch_mock)

    await touch(identity='id1', resource=peering_resource, settings=settings,
                namespace=peering_namespace)

    assert patch_mock.called
    patch = await patch_mock.call_args_list[0][0][0].json()
    assert set(patch['status']) == {'id1'}
    assert patch['status']['id1']['priority'] == 0
    assert patch['status']['id1']['lastseen'] == '2020-12-31T23:59:59.123456+00:00'
    assert patch['status']['id1']['lifetime'] == 60


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_expiring_a_peer_purges_it(
        hostname, aresponses, resp_mocker, settings,
        peering_resource, peering_namespace):

    settings.peering.name = 'name0'
    patch_mock = resp_mocker(return_value=aiohttp.web.json_response({}))
    url = peering_resource.get_url(name='name0', namespace=peering_namespace)
    aresponses.add(hostname, url, 'patch', patch_mock)

    await touch(identity='id1', resource=peering_resource, settings=settings,
                namespace=peering_namespace, lifetime=0)

    assert patch_mock.called
    patch = await patch_mock.call_args_list[0][0][0].json()
    assert set(patch['status']) == {'id1'}
    assert patch['status']['id1'] is None


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_logs_are_skipped_in_stealth_mode(
        hostname, aresponses, resp_mocker, settings, assert_logs,
        peering_resource, peering_namespace):

    settings.peering.stealth = True
    settings.peering.name = 'name0'
    patch_mock = resp_mocker(return_value=aiohttp.web.json_response({}))
    url = peering_resource.get_url(name='name0', namespace=peering_namespace)
    aresponses.add(hostname, url, 'patch', patch_mock)

    await touch(identity='id1', resource=peering_resource, settings=settings,
                namespace=peering_namespace)

    assert_logs(prohibited=[
        "Keep-alive in",
    ])


async def test_logs_are_logged_in_exposed_mode(
        hostname, aresponses, resp_mocker, settings, assert_logs,
        peering_resource, peering_namespace):

    settings.peering.stealth = False
    settings.peering.name = 'name0'
    patch_mock = resp_mocker(return_value=aiohttp.web.json_response({}))
    url = peering_resource.get_url(name='name0', namespace=peering_namespace)
    aresponses.add(hostname, url, 'patch', patch_mock)

    await touch(identity='id1', resource=peering_resource, settings=settings,
                namespace=peering_namespace)

    assert_logs([
        r"Keep-alive in 'name0' (in 'ns'|cluster-wide): ok",
    ])


@pytest.mark.parametrize('stealth', [True, False], ids=['stealth', 'exposed'])
async def test_logs_are_logged_when_absent(
        hostname, aresponses, resp_mocker, stealth, settings, assert_logs,
        peering_resource, peering_namespace):

    settings.peering.stealth = stealth
    settings.peering.name = 'name0'
    patch_mock = resp_mocker(return_value=aresponses.Response(status=404, reason='oops'))
    url = peering_resource.get_url(name='name0', namespace=peering_namespace)
    aresponses.add(hostname, url, 'patch', patch_mock)

    await touch(identity='id1', resource=peering_resource, settings=settings,
                namespace=peering_namespace)

    assert_logs([
        r"Keep-alive in 'name0' (in 'ns'|cluster-wide): not found",
    ])



import functools
import logging

import pytest

from kopf._core.actions.loggers import ObjectLogger
from kopf._core.engines.posting import event, exception, info, warn

OBJ1 = {'apiVersion': 'group1/version1', 'kind': 'Kind1',
        'metadata': {'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1'}}
REF1 = {'apiVersion': 'group1/version1', 'kind': 'Kind1',
        'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1'}


@pytest.fixture(autouse=True)
def _settings_via_contextvar(settings_via_contextvar):
    pass


async def test_loggers_setting_default_is_true(settings):
    assert settings.posting.loggers  # for backwards compatibility


async def test_loggers_setting_true_posts_loggers(settings, event_queue, event_queue_loop, caplog):
    settings.posting.loggers = True
    logger = ObjectLogger(body=OBJ1, settings=settings)
    with caplog.at_level(logging.INFO):
        logger.info("hello world")

    assert event_queue.qsize() == 1
    event1 = event_queue.get_nowait()
    assert event1.message == "hello world"


async def test_loggers_setting_false_skips_loggers(settings, event_queue, event_queue_loop, caplog):
    settings.posting.loggers = False
    logger = ObjectLogger(body=OBJ1, settings=settings)
    with caplog.at_level(logging.INFO):
        logger.info("this log should be skipped")

    assert event_queue.qsize() == 0


@pytest.mark.parametrize('routine', [
    functools.partial(event, type='irrelevant'),
    info,
    warn,
    exception,
])
@pytest.mark.parametrize('flag', [True, False], ids=['on', 'off'])
async def test_explicit_events_posted_regardless_of_settings(settings, event_queue, event_queue_loop, flag, routine):
    settings.posting.loggers = flag
    routine(OBJ1, reason="ExplicitReason", message="this should be posted")

    assert event_queue.qsize() == 1
    event1 = event_queue.get_nowait()
    assert event1.reason == "ExplicitReason"
    assert event1.message == "this should be posted"



import contextlib
from unittest.mock import Mock

import pytest

import kopf
from kopf._cogs.aiokits.aiovalues import Container
from kopf._core.engines.admission import admission_webhook_server


async def webhookfn(*_, **__):
    pass


async def test_requires_webserver_if_webhooks_are_defined(
        settings, registry, insights, resource):

    @kopf.on.validate(*resource, registry=registry)
    def fn_v(**_): pass

    @kopf.on.mutate(*resource, registry=registry)
    def fn_m(**_): pass

    container = Container()
    with pytest.raises(Exception) as err:
        settings.admission.server = None
        await admission_webhook_server(
            registry=registry,
            settings=settings,
            insights=insights,
            container=container,
            webhookfn=webhookfn,
        )

    assert "Admission handlers exist, but no admission server/tunnel" in str(err.value)


async def test_configures_client_configs(
        settings, registry, insights):

    async def server(_):
        yield {'url': 'https://hostname/'}

    container = Container()
    settings.admission.server = server
    await admission_webhook_server(
        registry=registry,
        settings=settings,
        insights=insights,
        container=container,
        webhookfn=webhookfn,
    )

    assert container.get_nowait() == {'url': 'https://hostname/'}


async def test_contextmanager_class(
        settings, registry, insights):

    aenter_mock = Mock()
    aexit_mock = Mock()
    container = Container()

    class CtxMgrServer:
        async def __aenter__(self) -> "CtxMgrServer":
            aenter_mock()
            return self

        async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
            aexit_mock()

        async def __call__(self, _):
            yield {'url': 'https://hostname/'}

    settings.admission.server = CtxMgrServer()
    await admission_webhook_server(
        registry=registry,
        settings=settings,
        insights=insights,
        container=container,
        webhookfn=webhookfn,
    )

    assert aenter_mock.call_count == 1
    assert aexit_mock.call_count == 1
    assert container.get_nowait() == {'url': 'https://hostname/'}


async def test_contextmanager_decorator(
        settings, registry, insights):

    aenter_mock = Mock()
    aexit_mock = Mock()
    container = Container()

    async def server(_):
        yield {'url': 'https://hostname/'}

    @contextlib.asynccontextmanager
    async def server_manager():
        aenter_mock()
        try:
            yield server
        finally:
            aexit_mock()

    settings.admission.server = server_manager()  # can be entered only once
    await admission_webhook_server(
        registry=registry,
        settings=settings,
        insights=insights,
        container=container,
        webhookfn=webhookfn,
    )

    assert aenter_mock.call_count == 1
    assert aexit_mock.call_count == 1
    assert container.get_nowait() == {'url': 'https://hostname/'}



"""
Ensure that the framework properly invokes or ignores the handlers
depending on the consistency or inconsistency of the incoming stream of events.
"""
import asyncio
import contextlib

import pytest

import kopf
from kopf._cogs.structs.ephemera import Memo
from kopf._core.engines.indexing import OperatorIndexers
from kopf._core.intents.causes import HANDLER_REASONS, Reason
from kopf._core.reactor.inventory import ResourceMemories
from kopf._core.reactor.processing import process_resource_event


@pytest.fixture(autouse=True)
def _no_throttling(mocker, resource, registry):
    @contextlib.asynccontextmanager
    async def not_throttled(*_, **__):
        yield True

    mocker.patch('kopf._core.actions.throttlers.throttled', not_throttled)


@pytest.mark.parametrize('cause_reason', HANDLER_REASONS)
async def test_preexisting_consistency(
        resource, registry, settings, handlers, cause_mock, cause_reason, k8s_mocked, looptime):
    event_type = None if cause_reason == Reason.RESUME else 'irrelevant'
    cause_mock.reason = cause_reason

    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=event_queue,
        consistency_time=None,  # assume the pre-existing consistency
    )

    assert looptime == 0
    assert handlers.event_mock.call_count == 1
    assert handlers.index_mock.call_count == 1
    assert handlers.create_mock.call_count == (1 if cause_reason == Reason.CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_reason == Reason.UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_reason == Reason.DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_reason == Reason.RESUME else 0)

    changing_mock = (
        handlers.create_mock if cause_reason == Reason.CREATE else
        handlers.update_mock if cause_reason == Reason.UPDATE else
        handlers.delete_mock if cause_reason == Reason.DELETE else
        handlers.resume_mock if cause_reason == Reason.RESUME else
        None  # and fail
    )
    assert handlers.event_mock.call_args_list[0][1]['_time'] == 0  # called instantly
    assert handlers.index_mock.call_args_list[0][1]['_time'] == 0  # called instantly
    assert changing_mock.call_args_list[0][1]['_time'] == 0  # called instantly


@pytest.mark.parametrize('cause_reason', HANDLER_REASONS)
async def test_past_consistency(
        resource, registry, settings, handlers, cause_mock, cause_reason, k8s_mocked, looptime):
    event_type = None if cause_reason == Reason.RESUME else 'irrelevant'
    cause_mock.reason = cause_reason

    await asyncio.sleep(100)  # fast-forward to the future
    assert looptime == 100

    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=event_queue,
        consistency_time=50,  # expect the consistency in the past
    )

    assert looptime == 100
    assert handlers.event_mock.call_count == 1
    assert handlers.index_mock.call_count == 1
    assert handlers.create_mock.call_count == (1 if cause_reason == Reason.CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_reason == Reason.UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_reason == Reason.DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_reason == Reason.RESUME else 0)

    changing_mock = (
        handlers.create_mock if cause_reason == Reason.CREATE else
        handlers.update_mock if cause_reason == Reason.UPDATE else
        handlers.delete_mock if cause_reason == Reason.DELETE else
        handlers.resume_mock if cause_reason == Reason.RESUME else
        None  # and fail
    )
    assert handlers.event_mock.call_args_list[0][1]['_time'] == 100  # called instantly
    assert handlers.index_mock.call_args_list[0][1]['_time'] == 100  # called instantly
    assert changing_mock.call_args_list[0][1]['_time'] == 100  # called instantly


@pytest.mark.parametrize('cause_reason', HANDLER_REASONS)
async def test_future_consistency(
        resource, registry, settings, handlers, cause_mock, cause_reason, k8s_mocked, looptime):
    settings.persistence.consistency_timeout = 55

    event_type = None if cause_reason == Reason.RESUME else 'irrelevant'
    cause_mock.reason = cause_reason

    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=event_queue,
        consistency_time=123,  # expect the consistency in the future
    )

    assert looptime == 123  # non-zero means it has slept for some time
    assert handlers.event_mock.call_count == 1
    assert handlers.index_mock.call_count == 1
    assert handlers.create_mock.call_count == (1 if cause_reason == Reason.CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_reason == Reason.UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_reason == Reason.DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_reason == Reason.RESUME else 0)

    changing_mock = (
        handlers.create_mock if cause_reason == Reason.CREATE else
        handlers.update_mock if cause_reason == Reason.UPDATE else
        handlers.delete_mock if cause_reason == Reason.DELETE else
        handlers.resume_mock if cause_reason == Reason.RESUME else
        None  # and fail
    )
    assert handlers.event_mock.call_args_list[0][1]['_time'] == 0  # called instantly
    assert handlers.index_mock.call_args_list[0][1]['_time'] == 0  # called instantly
    assert changing_mock.call_args_list[0][1]['_time'] == 123  # called after the deemed consistency


@pytest.mark.parametrize('cause_reason', HANDLER_REASONS)
async def test_no_consistency_wait_without_changing_handlers(
        resource, registry, settings, cause_mock, cause_reason, watching_handlers,
        k8s_mocked, looptime):
    cause_mock.reason = cause_reason

    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': '...', 'object': {}},
        event_queue=event_queue,
        consistency_time=456,  # expect the consistency in the future
    )

    assert looptime == 0  # zero means it did not sleep, as expected
    assert watching_handlers.event_mock.call_count == 1
    assert watching_handlers.index_mock.call_count == 1
    assert watching_handlers.event_mock.call_args_list[0][1]['_time'] == 0  # called instantly
    assert watching_handlers.index_mock.call_args_list[0][1]['_time'] == 0  # called instantly


@pytest.mark.parametrize('cause_reason', HANDLER_REASONS)
async def test_stream_pressure_awakening_prevents_change_handlers(
        resource, registry, settings, cause_mock, cause_reason, handlers,
        k8s_mocked, looptime):
    cause_mock.reason = cause_reason

    stream_pressure = asyncio.Event()
    asyncio.get_running_loop().call_later(123, stream_pressure.set)

    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': '...', 'object': {}},
        event_queue=event_queue,
        consistency_time=456,  # expect the consistency in the future
        stream_pressure=stream_pressure,
    )

    assert looptime == 123  # means: it has slept until awakened, but not until consistency
    assert handlers.event_mock.call_count == 1
    assert handlers.index_mock.call_count == 1
    assert handlers.event_mock.call_args_list[0][1]['_time'] == 0  # called instantly
    assert handlers.index_mock.call_args_list[0][1]['_time'] == 0  # called instantly

    changing_mock = (
        handlers.create_mock if cause_reason == Reason.CREATE else
        handlers.update_mock if cause_reason == Reason.UPDATE else
        handlers.delete_mock if cause_reason == Reason.DELETE else
        handlers.resume_mock if cause_reason == Reason.RESUME else
        None  # and fail
    )
    assert changing_mock.call_count == 0  # zero means the state-dependent handler was skipped


@pytest.mark.parametrize('cause_reason', HANDLER_REASONS)
async def test_resource_version_is_returned_from_patching(
        resource, registry, settings, cause_mock, cause_reason, watching_handlers,
        k8s_mocked, looptime):
    cause_mock.reason = Reason.CREATE

    k8s_mocked.patch.return_value = {'metadata': {'resourceVersion': 'some-rv'}}
    watching_handlers.event_mock.return_value = 'something-to-provoke-patching'

    event_queue = asyncio.Queue()
    rv = await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': '...', 'object': {}},
        event_queue=event_queue,
    )
    assert rv == 'some-rv'
    assert looptime == 0  # zero means everything has happened instantly



"""
Callback signatures for typing.

Since these signatures contain a lot of copy-pasted kwargs and are
not so important for the codebase, they are moved to this separate module.

As a rule of thumb, for every kwarg named ``whatever``, there should be
a corresponding type or class ``kopf.Whatever`` with all the typing tricks
(unions, optionals, partial ``Any`` values, etc) included.
"""
import datetime
from collections.abc import Callable, Collection
from typing import TYPE_CHECKING, Any, TypeVar

from kopf._cogs.configs import configuration
from kopf._cogs.helpers import typedefs
from kopf._cogs.structs import bodies, diffs, ephemera, patches, references, reviews
from kopf._core.actions import invocation
from kopf._core.intents import stoppers

if not TYPE_CHECKING:  # pragma: nocover
    # Define unspecified protocols for the runtime annotations -- to avoid "quoting".
    ActivityFn = Callable[..., invocation.SyncOrAsync[object | None]]
    IndexingFn = Callable[..., invocation.SyncOrAsync[object | None]]
    WatchingFn = Callable[..., invocation.SyncOrAsync[object | None]]
    ChangingFn = Callable[..., invocation.SyncOrAsync[object | None]]
    WebhookFn = Callable[..., invocation.SyncOrAsync[object | None]]
    DaemonFn = Callable[..., invocation.SyncOrAsync[object | None]]
    TimerFn = Callable[..., invocation.SyncOrAsync[object | None]]
    DelayFn = Callable[..., float]  # strictly sync, no async!
    WhenFilterFn = Callable[..., bool]  # strictly sync, no async!
    MetaFilterFn = Callable[..., bool]  # strictly sync, no async!
else:
    from mypy_extensions import Arg, DefaultNamedArg, KwArg, NamedArg

    # TODO:1: Split to specialised LoginFn, ProbeFn, StartupFn, etc. -- with different result types.
    # TODO:2: Try using ParamSpec to support index type checking in callbacks
    #         when PEP 612 is released (https://www.python.org/dev/peps/pep-0612/)
    ActivityFn = Callable[
        [
            NamedArg(configuration.OperatorSettings, "settings"),
            NamedArg(ephemera.Index[Any, Any], "*"),
            NamedArg(int, "retry"),
            NamedArg(datetime.datetime, "started"),
            NamedArg(datetime.timedelta, "runtime"),
            NamedArg(typedefs.Logger, "logger"),
            NamedArg(Any, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        invocation.SyncOrAsync[object | None]
    ]

    IndexingFn = Callable[
        [
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(str | None, "uid"),
            NamedArg(str | None, "name"),
            NamedArg(str | None, "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(typedefs.Logger, "logger"),
            NamedArg(Any, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        invocation.SyncOrAsync[object | None]
    ]

    WatchingFn = Callable[
        [
            NamedArg(str, "type"),
            NamedArg(bodies.RawEvent, "event"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(str | None, "uid"),
            NamedArg(str | None, "name"),
            NamedArg(str | None, "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(typedefs.Logger, "logger"),
            NamedArg(Any, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        invocation.SyncOrAsync[object | None]
    ]

    ChangingFn = Callable[
        [
            NamedArg(int, "retry"),
            NamedArg(datetime.datetime, "started"),
            NamedArg(datetime.timedelta, "runtime"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(str | None, "uid"),
            NamedArg(str | None, "name"),
            NamedArg(str | None, "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(str, "reason"),
            NamedArg(diffs.Diff, "diff"),
            NamedArg(bodies.BodyEssence | Any | None, "old"),
            NamedArg(bodies.BodyEssence | Any | None, "new"),
            NamedArg(typedefs.Logger, "logger"),
            NamedArg(Any, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        invocation.SyncOrAsync[object | None]
    ]

    WebhookFn = Callable[
        [
            NamedArg(bool, "dryrun"),
            NamedArg(list[str], "warnings"),  # mutable!
            NamedArg(str | None, "subresource"),
            NamedArg(reviews.UserInfo, "userinfo"),
            NamedArg(reviews.SSLPeer, "sslpeer"),
            NamedArg(reviews.Headers, "headers"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(str | None, "uid"),
            NamedArg(str | None, "name"),
            NamedArg(str | None, "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(typedefs.Logger, "logger"),
            NamedArg(Any, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        invocation.SyncOrAsync[object | None]
    ]

    DaemonFn = Callable[
        [
            NamedArg(stoppers.DaemonStopped, "stopped"),
            NamedArg(int, "retry"),
            NamedArg(datetime.datetime, "started"),
            NamedArg(datetime.timedelta, "runtime"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(str | None, "uid"),
            NamedArg(str | None, "name"),
            NamedArg(str | None, "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(typedefs.Logger, "logger"),
            NamedArg(Any, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        invocation.SyncOrAsync[object | None]
    ]

    TimerFn = Callable[
        [
            NamedArg(ephemera.Index[Any, Any], "*"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(str | None, "uid"),
            NamedArg(str | None, "name"),
            NamedArg(str | None, "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(typedefs.Logger, "logger"),
            NamedArg(Any, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        invocation.SyncOrAsync[object | None]
    ]

    DelayFn = Callable[
        [
            NamedArg(ephemera.Index[Any, Any], "*"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(str | None, "uid"),
            NamedArg(str | None, "name"),
            NamedArg(str | None, "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(typedefs.Logger, "logger"),
            NamedArg(Any, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        float  # strictly sync, no async!
    ]

    WhenFilterFn = Callable[
        [
            NamedArg(str, "type"),
            NamedArg(bodies.RawEvent, "event"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(str | None, "uid"),
            NamedArg(str | None, "name"),
            NamedArg(str | None, "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(diffs.Diff, "diff"),
            NamedArg(bodies.BodyEssence | Any | None, "old"),
            NamedArg(bodies.BodyEssence | Any | None, "new"),
            NamedArg(typedefs.Logger, "logger"),
            NamedArg(Any, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        bool  # strictly sync, no async!
    ]

    MetaFilterFn = Callable[
        [
            Arg(Any, "value"),
            NamedArg(str, "type"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(str | None, "uid"),
            NamedArg(str | None, "name"),
            NamedArg(str | None, "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(typedefs.Logger, "logger"),
            NamedArg(Any, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        bool  # strictly sync, no async!
    ]

SpawningFn = DaemonFn | TimerFn
_FnT = TypeVar('_FnT', WhenFilterFn, MetaFilterFn)


def not_(fn: _FnT) -> _FnT:
    def not_fn(*args: Any, **kwargs: Any) -> bool:
        return not fn(*args, **kwargs)
    return not_fn


def all_(fns: Collection[_FnT]) -> _FnT:
    def all_fn(*args: Any, **kwargs: Any) -> bool:
        return all(fn(*args, **kwargs) for fn in fns)
    return all_fn


def any_(fns: Collection[_FnT]) -> _FnT:
    def any_fn(*args: Any, **kwargs: Any) -> bool:
        return any(fn(*args, **kwargs) for fn in fns)
    return any_fn


def none_(fns: Collection[_FnT]) -> _FnT:
    def none_fn(*args: Any, **kwargs: Any) -> bool:
        return not any(fn(*args, **kwargs) for fn in fns)
    return none_fn



import datetime

import freezegun
import iso8601

from kopf._core.engines.peering import Peer


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_defaults():
    peer = Peer(identity='id')
    assert peer.identity == 'id'
    assert peer.lifetime == datetime.timedelta(seconds=60)
    assert peer.lastseen == iso8601.parse_date('2020-12-31T23:59:59.123456')


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_repr():
    peer = Peer(identity='some-id')
    text = repr(peer)
    assert text == "<Peer some-id: priority=0, lifetime=60, lastseen='2020-12-31T23:59:59.123456+00:00'>"


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_priority_specified():
    peer = Peer(identity='id', priority=123)
    assert peer.priority == 123


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_priority_unspecified():
    peer = Peer(identity='id')
    assert peer.priority == 0


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_with_lifetime_as_number():
    peer = Peer(identity='id', lifetime=123)
    assert peer.lifetime == datetime.timedelta(seconds=123)


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_with_lifetime_unspecified():
    peer = Peer(identity='id')
    assert peer.lifetime == datetime.timedelta(seconds=60)


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_with_lastseen_as_string():
    peer = Peer(identity='id', lastseen='2020-01-01T12:34:56.789123')
    assert peer.lastseen == iso8601.parse_date('2020-01-01T12:34:56.789123')


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_with_lastseen_unspecified():
    peer = Peer(identity='id')
    assert peer.lastseen == iso8601.parse_date('2020-12-31T23:59:59.123456')


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_as_alive():
    peer = Peer(
        identity='id',
        lifetime=10,
        lastseen='2020-12-31T23:59:50.123456',  # less than 10 seconds before "now"
    )
    assert peer.lifetime == datetime.timedelta(seconds=10)
    assert peer.lastseen == iso8601.parse_date('2020-12-31T23:59:50.123456')
    assert peer.deadline == iso8601.parse_date('2021-01-01T00:00:00.123456')
    assert peer.is_dead is False


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_as_dead():
    peer = Peer(
        identity='id',
        lifetime=10,
        lastseen='2020-12-31T23:59:49.123456',  # 10 seconds before "now" sharp
    )
    assert peer.lifetime == datetime.timedelta(seconds=10)
    assert peer.lastseen == iso8601.parse_date('2020-12-31T23:59:49.123456')
    assert peer.deadline == iso8601.parse_date('2020-12-31T23:59:59.123456')
    assert peer.is_dead is True



import collections.abc

import pytest

import kopf
from kopf._cogs.structs.references import Selector
from kopf._core.intents.causes import Activity, ChangingCause, WatchingCause


# Used in the tests. Must be global-scoped, or its qualname will be affected.
def some_fn():
    pass


def test_generic_registry_via_iter(
        generic_registry_cls, cause_factory):

    cause = cause_factory(generic_registry_cls)
    registry = generic_registry_cls()
    iterator = registry.iter_handlers(cause)

    assert isinstance(iterator, collections.abc.Iterator)
    assert not isinstance(iterator, collections.abc.Collection)
    assert not isinstance(iterator, collections.abc.Container)
    assert not isinstance(iterator, (list, tuple))

    handlers = list(iterator)
    assert not handlers


def test_generic_registry_via_list(
        generic_registry_cls, cause_factory):

    cause = cause_factory(generic_registry_cls)
    registry = generic_registry_cls()
    handlers = registry.get_handlers(cause)

    assert isinstance(handlers, collections.abc.Iterable)
    assert isinstance(handlers, collections.abc.Container)
    assert isinstance(handlers, collections.abc.Collection)
    assert not handlers


@pytest.mark.parametrize('activity', list(Activity))
def test_operator_registry_with_activity_via_iter(
        operator_registry_cls, activity):

    registry = operator_registry_cls()
    iterator = registry._activities.iter_handlers(activity=activity)

    assert isinstance(iterator, collections.abc.Iterator)
    assert not isinstance(iterator, collections.abc.Collection)
    assert not isinstance(iterator, collections.abc.Container)
    assert not isinstance(iterator, (list, tuple))

    handlers = list(iterator)
    assert not handlers


def test_operator_registry_watching_handlers_via_iter(
        operator_registry_cls, cause_factory):

    cause = cause_factory(WatchingCause)
    registry = operator_registry_cls()
    iterator = registry._watching.iter_handlers(cause)

    assert isinstance(iterator, collections.abc.Iterator)
    assert not isinstance(iterator, collections.abc.Collection)
    assert not isinstance(iterator, collections.abc.Container)
    assert not isinstance(iterator, (list, tuple))

    handlers = list(iterator)
    assert not handlers


def test_operator_registry_changing_handlers_via_iter(
        operator_registry_cls, cause_factory):

    cause = cause_factory(ChangingCause)
    registry = operator_registry_cls()
    iterator = registry._changing.iter_handlers(cause)

    assert isinstance(iterator, collections.abc.Iterator)
    assert not isinstance(iterator, collections.abc.Collection)
    assert not isinstance(iterator, collections.abc.Container)
    assert not isinstance(iterator, (list, tuple))

    handlers = list(iterator)
    assert not handlers


@pytest.mark.parametrize('activity', list(Activity))
def test_operator_registry_with_activity_via_list(
        operator_registry_cls, activity):

    registry = operator_registry_cls()
    handlers = registry._activities.get_handlers(activity=activity)

    assert isinstance(handlers, collections.abc.Iterable)
    assert isinstance(handlers, collections.abc.Container)
    assert isinstance(handlers, collections.abc.Collection)
    assert not handlers


def test_operator_registry_watching_handlers_via_list(
        operator_registry_cls, cause_factory):

    cause = cause_factory(WatchingCause)
    registry = operator_registry_cls()
    handlers = registry._watching.get_handlers(cause)

    assert isinstance(handlers, collections.abc.Iterable)
    assert isinstance(handlers, collections.abc.Container)
    assert isinstance(handlers, collections.abc.Collection)
    assert not handlers


def test_operator_registry_changing_handlers_via_list(
        operator_registry_cls, cause_factory):

    cause = cause_factory(ChangingCause)
    registry = operator_registry_cls()
    handlers = registry._changing.get_handlers(cause)

    assert isinstance(handlers, collections.abc.Iterable)
    assert isinstance(handlers, collections.abc.Container)
    assert isinstance(handlers, collections.abc.Collection)
    assert not handlers


def test_all_handlers(operator_registry_cls):
    registry = operator_registry_cls()

    @kopf.index('resource', registry=registry)
    @kopf.on.event('resource', registry=registry)
    @kopf.on.event('resource', registry=registry)
    @kopf.on.create('resource', registry=registry)
    @kopf.on.update('resource', registry=registry)
    @kopf.on.delete('resource', registry=registry)
    @kopf.on.resume('resource', registry=registry)
    @kopf.on.timer('resource', registry=registry)
    @kopf.on.daemon('resource', registry=registry)
    def fn(**_): pass

    assert len(registry._indexing.get_all_handlers()) == 1
    assert len(registry._watching.get_all_handlers()) == 2
    assert len(registry._spawning.get_all_handlers()) == 2
    assert len(registry._changing.get_all_handlers()) == 4


def test_all_selectors(operator_registry_cls):
    registry = operator_registry_cls()

    @kopf.index('resource0', registry=registry)
    @kopf.on.event('resource1', registry=registry)
    @kopf.on.event('resource2', registry=registry)
    @kopf.on.create('resource1', registry=registry)
    @kopf.on.update('resource2', registry=registry)
    @kopf.on.delete('resource3', registry=registry)
    @kopf.on.resume('resource4', registry=registry)
    @kopf.on.timer('resource5', registry=registry)
    @kopf.on.daemon('resource6', registry=registry)
    def fn(**_): pass

    assert registry._indexing.get_all_selectors() == {Selector('resource0')}
    assert registry._watching.get_all_selectors() == {Selector('resource1'), Selector('resource2')}
    assert registry._spawning.get_all_selectors() == {Selector('resource5'), Selector('resource6')}
    assert registry._changing.get_all_selectors() == {Selector('resource1'), Selector('resource2'),
                                                      Selector('resource3'), Selector('resource4')}



import asyncio
from collections.abc import Callable, Collection, Iterable, Iterator


class Toggle:
    """
    An synchronisation primitive that can be awaited both until set or cleared.

    For one-directional toggles, ``asyncio.Event`` is sufficient.
    But these events cannot be awaited until cleared.

    The bi-directional toggles are needed in some places in the code, such as
    in the population/depletion of :class:`Vault`, or as in the operator pause.

    The optional name is used only for hinting in reprs. It can be used when
    there are many toggles, and they need to be distinguished somehow.
    """

    def __init__(
            self,
            __state: bool = False,
            *,
            name: str | None = None,
            condition: asyncio.Condition | None = None,
    ) -> None:
        super().__init__()
        self._condition = condition if condition is not None else asyncio.Condition()
        self._state: bool = bool(__state)
        self._name = name

    def __repr__(self) -> str:
        clsname = self.__class__.__name__
        toggled = 'on' if self._state else 'off'
        if self._name is None:
            return f'<{clsname}: {toggled}>'
        else:
            return f'<{clsname}: {self._name}: {toggled}>'

    def __bool__(self) -> bool:
        raise NotImplementedError  # to protect against accidental misuse

    def is_on(self) -> bool:
        return self._state

    def is_off(self) -> bool:
        return not self._state

    async def turn_to(self, __state: bool) -> None:
        """ Turn the toggle on/off, and wake up the tasks waiting for that. """
        async with self._condition:
            self._state = bool(__state)
            self._condition.notify_all()

    async def wait_for(self, __state: bool) -> None:
        """ Wait until the toggle is turned on/off as expected (if not yet). """
        async with self._condition:
            await self._condition.wait_for(lambda: self._state == bool(__state))

    @property
    def name(self) -> str | None:
        return self._name


class ToggleSet(Collection[Toggle]):
    """
    A read-only checker for multiple toggles.

    The toggle-checker does not have its own state to be turned on/off.

    The positional argument is a function, usually :func:`any` or :func:`all`,
    which takes an iterable of all individual toggles' states (on/off),
    and calculates the overall state of the toggle set.

    With :func:`any`, the set is "on" when at least one child toggle is "on"
    (and it has at least one child), and it is "off" when all children toggles
    are "off" (or if it has no children toggles at all).

    With :func:`all`, the set is "on" when all of its children toggles are "on"
    (or it has no children at all), and it is "off" when at least one child
    toggle is "off" (and there is at least one toggle).

    The multi-toggle sets are used mostly for operator pausing,
    e.g. in peering and in index pre-population. For a practical example,
    in peering, every individual peering identified by name and namespace has
    its own individual toggle to manage, but the whole set of toggles of all
    names & namespaces is used for pausing the operator as one single toggle.
    In index pre-population, the toggles are used on the operator's startup
    to temporarily delay the actual resource handling until all index-handlers
    of all involved resources and resource kinds are processed and stored.

    Note: the set can only contain toggles that were produced by the set;
    externally produced toggles cannot be added, since they do not share
    the same condition object, which is used for synchronisation/notifications.
    """

    def __init__(self, fn: Callable[[Iterable[bool]], bool]) -> None:
        super().__init__()
        self._condition = asyncio.Condition()
        self._toggles: set[Toggle] = set()
        self._fn = fn

    def __repr__(self) -> str:
        return repr(self._toggles)

    def __len__(self) -> int:
        return len(self._toggles)

    def __iter__(self) -> Iterator[Toggle]:
        return iter(self._toggles)

    def __contains__(self, toggle: object) -> bool:
        return toggle in self._toggles

    def __bool__(self) -> bool:
        raise NotImplementedError  # to protect against accidental misuse

    def is_on(self) -> bool:
        return self._fn(toggle.is_on() for toggle in self._toggles)

    def is_off(self) -> bool:
        return not self.is_on()

    async def wait_for(self, __state: bool) -> None:
        async with self._condition:
            await self._condition.wait_for(lambda: self.is_on() == bool(__state))

    async def make_toggle(
            self,
            __val: bool = False,
            *,
            name: str | None = None,
    ) -> Toggle:
        toggle = Toggle(__val, name=name, condition=self._condition)
        async with self._condition:
            self._toggles.add(toggle)
            self._condition.notify_all()
        return toggle

    async def drop_toggle(self, toggle: Toggle) -> None:
        async with self._condition:
            self._toggles.discard(toggle)
            self._condition.notify_all()

    async def drop_toggles(self, toggles: Iterable[Toggle]) -> None:
        async with self._condition:
            self._toggles.difference_update(toggles)
            self._condition.notify_all()



import asyncio

import pytest

from kopf._core.engines.posting import event_queue_loop_var, event_queue_var


@pytest.fixture(autouse=True)
def _caplog_all_levels(caplog):
    caplog.set_level(0)


# We can get a properly scoped running loop of the test only in the async fixture.
@pytest.fixture
async def _event_queue_running_loop():
    return asyncio.get_running_loop()


@pytest.fixture(autouse=True)
def event_queue_loop(_event_queue_running_loop):  # must be sync-def
    token = event_queue_loop_var.set(_event_queue_running_loop)
    try:
        yield
    finally:
        event_queue_loop_var.reset(token)


@pytest.fixture(autouse=True)
def event_queue():
    queue = asyncio.Queue()
    token = event_queue_var.set(queue)
    try:
        yield queue
    finally:
        event_queue_var.reset(token)



"""
The routines to manipulate the handler progression over the event cycle.

Used to track which handlers are finished, which are not yet,
and how many retries were there.

There could be more than one low-level k8s watch-events per one actual
high-level kopf-event (a cause). The handlers are called at different times,
and the overall handling routine should persist the handler status somewhere.

The states are persisted in a state storage:
see :mod:`kopf._cogs.configs.progress`.

MOCKED LOOP TIME:

**For testability,** we use ``basetime + timedelta(seconds=loop.time())``
to calculate the "now" moment instead of ``datetime.utcnow()``.

The "basetime" is an imaginary UTC time when the loop clock was zero (``0.0``)
and is calculated as ``datetime.utcnow() - timedelta(seconds=loop.time())``
(assuming these two calls are almost instant and the precision loss is low).

In normal run mode, the "basetime" remains constant for the entire life time
of an event loop, since both loop time and wall-clock time move forward with
the same speed: the calculation of "basetime" always produces the same result.

In test mode, the loop time is mocked and moves as events (e.g. sleeps) happen:
it can move (much) faster than the wall-clock time, e.g. 100s of loop seconds
in 1/100th of a wall-clock second; or it can freeze and not move at all.

PROBLEMATIC INACCURACY:

Because of a highly unprecise and everchanging component in the formula
of the "basetime" — the non-mockable UTC clock — the "basetime" calculation
can give different results at different times even if executed fast enough.

To reduce the inaccuracy introduced by sequential UTC time measurements,
we calculate the "basetime" once per every global state object created
and push it down to owned state objects of the individual handlers
in this halding cycle of this resource object in this unit-test.

That gives us sufficient accuracy while remaining simple enough, assuming that
there are no multiple concurrent global state objects per single unit-test.
_(An alternative would be to calculate the "basetime" on event loop creation
or to cache it per event loop in a global WeakDict, but that is an overkill.)_

SUFFICIENT ACCURACY:

With this approach and ``looptime``__, we can detach from the wall-clock time
in tests and simulate the time's rapid movement into the future by "recovering"
the "now" moment as ``basetime + timedelta(seconds=loop.time())`` (see above) —
without wall-clock delays or hitting the issues with code execution overhead.

Note that there is no UTC clock involved now, only the controled loop clock,
so multiple sequential calculation will lead to predictable abd precise results,
especially when the loop clock is frozen (i.e. constant for a short duration).

__ https://github.com/nolar/looptime

USER PERSPECTIVE:

This time math is never exposed to users and never persisted in storages.
It is used only internally to decouple the operator routines from the system
clock and strictly couple it to the time of the loop.

IMPLEMENTATION DETAILS:

Q: Why do we store UTC time in the fields instead of the floats with loop time?
A: If we store floats in the fields, we need to do the math on every
fetching/storing operation, which introduces minor divergence in supposedly
constant data as stored in the external storages. Instead, we only calculate
the "now" moment. As a result, the precision loss is seen only at runtime checks
and is indistinguishanle from the loop clock sensitivity.
"""
import asyncio
import collections.abc
import copy
import dataclasses
import datetime
from collections.abc import Collection, Iterable, Iterator, Mapping
from typing import Any, NamedTuple, overload

import iso8601

from kopf._cogs.configs import progress
from kopf._cogs.structs import bodies, ids, patches
from kopf._core.actions import execution


@dataclasses.dataclass(frozen=True)
class HandlerState(execution.HandlerState):
    """
    A persisted state of a single handler, as stored on the resource's status.

    Note the difference: :class:`Outcome` is for in-memory results of handlers,
    which is then additionally converted before being storing as a state.

    Active handler states are those used in .done/.delays for the current
    handling cycle & the current cause. Passive handler states are those
    carried over for logging of counts/extras, and for final state purging,
    but not participating in the current handling cycle.
    """

    active: bool  # whether it is used in done/delays [T] or only in counters/purges [F].
    basetime: datetime.datetime  # a moment when the loop time was zero
    started: datetime.datetime
    stopped: datetime.datetime | None = None  # None means it is still running (e.g. delayed).
    delayed: datetime.datetime | None = None  # None means it is finished (succeeded/failed).
    purpose: str | None = None  # None is a catch-all marker for upgrades/rollbacks.
    retries: int = 0
    success: bool = False
    failure: bool = False
    message: str | None = None
    subrefs: Collection[ids.HandlerId] = ()  # ids of actual sub-handlers of all levels deep.
    _origin: progress.ProgressRecord | None = None  # to check later if it has actually changed.

    @property
    def finished(self) -> bool:
        return bool(self.success or self.failure)

    @property
    def sleeping(self) -> bool:
        now = self.basetime + datetime.timedelta(seconds=asyncio.get_running_loop().time())
        return not self.finished and self.delayed is not None and self.delayed > now

    @property
    def awakened(self) -> bool:
        return bool(not self.finished and not self.sleeping)

    @property
    def runtime(self) -> datetime.timedelta:
        now = self.basetime + datetime.timedelta(seconds=asyncio.get_running_loop().time())
        return now - self.started

    @classmethod
    def from_scratch(
            cls,
            *,
            basetime: datetime.datetime,
            purpose: str | None = None,
    ) -> "HandlerState":
        now = basetime + datetime.timedelta(seconds=asyncio.get_running_loop().time())
        return cls(active=True, basetime=basetime, started=now, purpose=purpose)

    @classmethod
    def from_storage(
            cls,
            __d: progress.ProgressRecord,
            *,
            basetime: datetime.datetime,
    ) -> "HandlerState":
        now = basetime + datetime.timedelta(seconds=asyncio.get_running_loop().time())
        return cls(
            active=False,
            basetime=basetime,
            started=parse_iso8601(__d.get('started')) or now,
            stopped=parse_iso8601(__d.get('stopped')),
            delayed=parse_iso8601(__d.get('delayed')),
            purpose=__d.get('purpose') if __d.get('purpose') else None,
            retries=__d.get('retries') or 0,
            success=__d.get('success') or False,
            failure=__d.get('failure') or False,
            message=__d.get('message'),
            subrefs=__d.get('subrefs') or (),
            _origin=__d,
        )

    def for_storage(self) -> progress.ProgressRecord:
        return progress.ProgressRecord(
            started=None if self.started is None else format_iso8601(self.started),
            stopped=None if self.stopped is None else format_iso8601(self.stopped),
            delayed=None if self.delayed is None else format_iso8601(self.delayed),
            purpose=None if self.purpose is None else str(self.purpose),
            retries=None if self.retries is None else int(self.retries),
            success=None if self.success is None else bool(self.success),
            failure=None if self.failure is None else bool(self.failure),
            message=None if self.message is None else str(self.message),
            subrefs=None if not self.subrefs else list(sorted(self.subrefs)),
        )

    def as_in_storage(self) -> Mapping[str, Any]:
        # Nones are not stored by Kubernetes, so we filter them out for comparison.
        return {key: val for key, val in self.for_storage().items() if val is not None}

    def as_active(self) -> "HandlerState":
        return dataclasses.replace(self, active=True)

    def with_purpose(
            self,
            purpose: str | None,
    ) -> "HandlerState":
        return dataclasses.replace(self, purpose=purpose)

    def with_outcome(
            self,
            outcome: execution.Outcome,
    ) -> "HandlerState":
        now = self.basetime + datetime.timedelta(seconds=asyncio.get_running_loop().time())
        cls = type(self)
        return cls(
            active=self.active,
            basetime=self.basetime,
            purpose=self.purpose,
            started=self.started,
            stopped=self.stopped if self.stopped is not None else now if outcome.final else None,
            delayed=now + datetime.timedelta(seconds=outcome.delay) if outcome.delay is not None else None,
            success=bool(outcome.final and outcome.exception is None),
            failure=bool(outcome.final and outcome.exception is not None),
            retries=(self.retries if self.retries is not None else 0) + 1,
            message=None if outcome.exception is None else str(outcome.exception),
            subrefs=list(sorted(set(self.subrefs) | set(outcome.subrefs))),  # deduplicate
            _origin=self._origin,
        )


class StateCounters(NamedTuple):
    success: int
    failure: int
    running: int


class State(execution.State):
    """
    A state of selected handlers, as persisted in the object's status.

    The state consists of simple YAML-/JSON-serializable values only:
    string handler ids as the keys; strings, booleans, integers as the values.

    The state is immutable: once created, it cannot be changed, and does not
    reflect the changes in the object's status. A new state is created every
    time some changes/outcomes are merged into the current state.
    """
    _states: Mapping[ids.HandlerId, HandlerState]

    # Eliminate even the smallest microsecond-scale deviations by using the shared base time.
    # The deviations can come from UTC wall-clock time slowly moving during the run (CPU overhead).
    basetime: datetime.datetime

    def __init__(
            self,
            __src: Mapping[ids.HandlerId, HandlerState],
            *,
            basetime: datetime.datetime,
            purpose: str | None = None,
    ):
        super().__init__()
        self._states = dict(__src)
        self.purpose = purpose
        self.basetime = basetime

    @classmethod
    def from_scratch(cls) -> "State":
        return cls({}, basetime=_get_basetime())

    @classmethod
    def from_storage(
            cls,
            *,
            body: bodies.Body,
            storage: progress.ProgressStorage,
            handlers: Iterable[execution.Handler],
    ) -> "State":
        basetime = _get_basetime()
        handler_ids = {handler.id for handler in handlers}
        handler_states: dict[ids.HandlerId, HandlerState] = {}
        for handler_id in handler_ids:
            content = storage.fetch(key=handler_id, body=body)
            if content is not None:
                handler_states[handler_id] = HandlerState.from_storage(content, basetime=basetime)
        return cls(handler_states, basetime=basetime)

    def with_purpose(
            self,
            purpose: str | None,
            handlers: Iterable[execution.Handler] = (),  # to be re-purposed
    ) -> "State":
        handler_states: dict[ids.HandlerId, HandlerState] = dict(self)
        for handler in handlers:
            handler_states[handler.id] = handler_states[handler.id].with_purpose(purpose)
        cls = type(self)
        return cls(handler_states, basetime=self.basetime, purpose=purpose)

    def with_handlers(
            self,
            handlers: Iterable[execution.Handler],
    ) -> "State":
        handler_states: dict[ids.HandlerId, HandlerState] = dict(self)
        for handler in handlers:
            if handler.id not in handler_states:
                handler_states[handler.id] = HandlerState.from_scratch(
                    basetime=self.basetime, purpose=self.purpose)
            else:
                handler_states[handler.id] = handler_states[handler.id].as_active()
        cls = type(self)
        return cls(handler_states, basetime=self.basetime, purpose=self.purpose)

    def with_outcomes(
            self,
            outcomes: Mapping[ids.HandlerId, execution.Outcome],
    ) -> "State":
        unknown_ids = [handler_id for handler_id in outcomes if handler_id not in self]
        if unknown_ids:
            raise RuntimeError(f"Unexpected outcomes for unknown handlers: {unknown_ids!r}")

        cls = type(self)
        return cls({
            handler_id: (handler_state if handler_id not in outcomes else
                         handler_state.with_outcome(outcomes[handler_id]))
            for handler_id, handler_state in self._states.items()
        }, basetime=self.basetime, purpose=self.purpose)

    def without_successes(self) -> "State":
        cls = type(self)
        return cls({
            handler_id: handler_state
            for handler_id, handler_state in self._states.items()
            if not handler_state.success # i.e. failures & in-progress/retrying
        }, basetime=self.basetime)

    def store(
            self,
            body: bodies.Body,
            patch: patches.Patch,
            storage: progress.ProgressStorage,
    ) -> None:
        for handler_id, handler_state in self._states.items():
            full_record = handler_state.for_storage()
            pure_record = handler_state.as_in_storage()
            if pure_record != handler_state._origin:
                storage.store(key=handler_id, record=full_record, body=body, patch=patch)
        storage.flush()

    def purge(
            self,
            *,
            body: bodies.Body,
            patch: patches.Patch,
            storage: progress.ProgressStorage,
            handlers: Iterable[execution.Handler],
    ) -> None:
        # Purge only our own handlers and their direct & indirect sub-handlers of all levels deep.
        # Ignore other handlers (e.g. handlers of other operators).
        handler_ids = {handler.id for handler in handlers}
        for handler_id in handler_ids:
            storage.purge(key=handler_id, body=body, patch=patch)
        for handler_id, handler_state in self._states.items():
            if handler_id not in handler_ids:
                storage.purge(key=handler_id, body=body, patch=patch)
            for subref in handler_state.subrefs:
                storage.purge(key=subref, body=body, patch=patch)
        storage.flush()

    def __len__(self) -> int:
        return len(self._states)

    def __iter__(self) -> Iterator[ids.HandlerId]:
        return iter(self._states)

    def __getitem__(self, item: ids.HandlerId) -> HandlerState:
        return self._states[item]

    @property
    def done(self) -> bool:
        # In particular, no handlers means that it is "done" even before doing.
        return all(
            handler_state.finished for handler_state in self._states.values()
            if handler_state.active
        )

    @property
    def extras(self) -> Mapping[str, StateCounters]:
        purposes = {
            handler_state.purpose for handler_state in self._states.values()
            if handler_state.purpose is not None and handler_state.purpose != self.purpose
        }
        counters = {
            purpose: StateCounters(
                success=len([1 for handler_state in self._states.values()
                            if handler_state.purpose == purpose and handler_state.success]),
                failure=len([1 for handler_state in self._states.values()
                            if handler_state.purpose == purpose and handler_state.failure]),
                running=len([1 for handler_state in self._states.values()
                            if handler_state.purpose == purpose and not handler_state.finished]),
            )
            for purpose in purposes
        }
        return counters

    @property
    def counts(self) -> StateCounters:
        purposeful_states = [
            handler_state for handler_state in self._states.values()
            if self.purpose is None or handler_state.purpose is None
               or handler_state.purpose == self.purpose
        ]
        return StateCounters(
            success=len([1 for handler_state in purposeful_states if handler_state.success]),
            failure=len([1 for handler_state in purposeful_states if handler_state.failure]),
            running=len([1 for handler_state in purposeful_states if not handler_state.finished]),
        )

    @property
    def delay(self) -> float | None:
        delays = self.delays  # calculate only once, to save bit of CPU
        return min(delays) if delays else None

    @property
    def delays(self) -> Collection[float]:
        """
        Resulting delays for the handlers (only the postponed ones).

        The delays are then reduced to one single sleep in the top-level
        processing routine, based on all delays of different origin:
        e.g. postponed daemons, stopping daemons, temporarily failed handlers.
        """
        now = self.basetime + datetime.timedelta(seconds=asyncio.get_running_loop().time())
        return [
            max(0.0, (handler_state.delayed - now).total_seconds()) if handler_state.delayed else 0
            for handler_state in self._states.values()
            if handler_state.active and not handler_state.finished
        ]


def deliver_results(
        *,
        outcomes: Mapping[ids.HandlerId, execution.Outcome],
        patch: patches.Patch,
) -> None:
    """
    Store the results (as returned from the handlers) to the resource.

    This is not the handlers' state persistence, but the results' persistence.

    First, the state persistence is stored under ``.status.kopf.progress``,
    and can (later) be configured to be stored in different fields for different
    operators operating the same objects: ``.status.kopf.{somename}.progress``.
    The handlers' result are stored in the top-level ``.status``.

    Second, the handler results can (also later) be delivered to other objects,
    e.g. to their owners or label-selected related objects. For this, another
    class/module will be added.

    For now, we keep state- and result persistence in one module, but separated.
    """
    for handler_id, outcome in outcomes.items():
        if outcome.exception is not None:
            pass
        elif outcome.result is None:
            pass
        elif isinstance(outcome.result, collections.abc.Mapping):
            # TODO: merge recursively (patch-merge), do not overwrite the keys if they are present.
            patch.setdefault('status', {}).setdefault(handler_id, {}).update(outcome.result)
        else:
            patch.setdefault('status', {})[handler_id] = copy.deepcopy(outcome.result)


@overload
def format_iso8601(val: None) -> None:
    ...


@overload
def format_iso8601(val: datetime.datetime) -> str:
    ...


def format_iso8601(val: datetime.datetime | None) -> str | None:
    return None if val is None else val.isoformat(timespec='microseconds')


@overload
def parse_iso8601(val: None) -> None:
    ...


@overload
def parse_iso8601(val: str) -> datetime.datetime:
    ...


def parse_iso8601(val: str | None) -> datetime.datetime | None:
    return None if val is None else iso8601.parse_date(val, default_timezone=None)


def _get_basetime() -> datetime.datetime:
    loop = asyncio.get_running_loop()
    return datetime.datetime.now(tz=datetime.timezone.utc) - datetime.timedelta(seconds=loop.time())



import asyncio

import kopf


async def test_timer_regular_interval(
        resource, dummy, assert_logs, k8s_mocked, simulate_cycle, looptime):
    trigger = asyncio.Condition()

    @kopf.timer(*resource, id='fn', interval=1.0, sharp=False)
    async def fn(**kwargs):
        dummy.mock(**kwargs)
        async with trigger:
            trigger.notify_all()
        await asyncio.sleep(0.3)  # simulate a slow operation

    await simulate_cycle({})
    async with trigger:
        await trigger.wait()
        await trigger.wait()

    assert dummy.mock.call_count == 2
    assert looptime == 1.3


async def test_timer_sharp_interval(
        resource, dummy, assert_logs, k8s_mocked, simulate_cycle, looptime):
    trigger = asyncio.Condition()

    @kopf.timer(*resource, id='fn', interval=1.0, sharp=True)
    async def fn(**kwargs):
        dummy.mock(**kwargs)
        async with trigger:
            trigger.notify_all()
        await asyncio.sleep(0.3)  # simulate a slow operation

    await simulate_cycle({})
    async with trigger:
        await trigger.wait_for(lambda: dummy.mock.call_count >= 2)

    assert dummy.mock.call_count == 2
    assert looptime == 1  # not 1.3!



import json

import pytest

from kopf._cogs.configs.progress import AnnotationsProgressStorage, ProgressRecord, \
                                        ProgressStorage, SmartProgressStorage, \
                                        StatusProgressStorage
from kopf._cogs.structs.bodies import Body
from kopf._cogs.structs.ids import HandlerId
from kopf._cogs.structs.patches import Patch

ALL_STORAGES = [AnnotationsProgressStorage, StatusProgressStorage, SmartProgressStorage]
ANNOTATIONS_POPULATING_STORAGES = [AnnotationsProgressStorage, SmartProgressStorage]
STATUS_POPULATING_STORAGES = [StatusProgressStorage, SmartProgressStorage]

CONTENT_DATA_1 = ProgressRecord(
    started='2020-01-01T00:00:00',
    stopped='2020-12-31T23:59:59',
    delayed='3000-01-01T00:00:00',
    retries=123,
    success=False,
    failure=False,
    message=None,
    subrefs=None,
)

CONTENT_DATA_2 = ProgressRecord(
    started='2021-01-01T00:00:00',
    stopped='2021-12-31T23:59:59',
    delayed='3001-01-01T00:00:00',
    retries=456,
    success=False,
    failure=False,
    message="Some error.",
    subrefs=['sub1', 'sub2'],
)

CONTENT_JSON_1 = json.dumps(CONTENT_DATA_1, separators=(',', ':'))
CONTENT_JSON_2 = json.dumps(CONTENT_DATA_2, separators=(',', ':'))


#
# Storage creation.
#


def test_status_storage_with_defaults():
    storage = StatusProgressStorage()
    assert storage.field == ('status', 'kopf', 'progress')  # as before the change
    assert storage.touch_field == ('status', 'kopf', 'dummy')  # as before the change


def test_status_storage_with_name():
    storage = StatusProgressStorage(name='my-operator')
    assert storage.field == ('status', 'my-operator', 'progress')
    assert storage.touch_field == ('status', 'my-operator', 'dummy')


def test_status_storage_with_field():
    storage = StatusProgressStorage(field='status.my-operator', touch_field='status.my-dummy')
    assert storage.field == ('status', 'my-operator')
    assert storage.touch_field == ('status', 'my-dummy')


def test_annotations_storage_with_defaults():
    storage = AnnotationsProgressStorage()
    assert storage.prefix == 'kopf.zalando.org'


def test_annotations_storage_with_prefix():
    storage = AnnotationsProgressStorage(prefix='my-operator.my-company.com')
    assert storage.prefix == 'my-operator.my-company.com'


def test_smart_storage_with_defaults():
    storage = SmartProgressStorage()
    assert isinstance(storage.storages[0], AnnotationsProgressStorage)
    assert isinstance(storage.storages[1], StatusProgressStorage)
    assert storage.storages[0].prefix == 'kopf.zalando.org'
    assert storage.storages[1].field == ('status', 'kopf', 'progress')
    assert storage.storages[1].touch_field == ('status', 'kopf', 'dummy')


def test_smart_storage_with_name():
    storage = SmartProgressStorage(name='my-operator')
    assert isinstance(storage.storages[0], AnnotationsProgressStorage)
    assert isinstance(storage.storages[1], StatusProgressStorage)
    assert storage.storages[0].prefix == 'kopf.zalando.org'
    assert storage.storages[1].field == ('status', 'my-operator', 'progress')
    assert storage.storages[1].touch_field == ('status', 'my-operator', 'dummy')


def test_smart_storage_with_field():
    storage = SmartProgressStorage(field='status.my-operator', touch_field='status.my-dummy')
    assert isinstance(storage.storages[0], AnnotationsProgressStorage)
    assert isinstance(storage.storages[1], StatusProgressStorage)
    assert storage.storages[0].prefix == 'kopf.zalando.org'
    assert storage.storages[1].field == ('status', 'my-operator')
    assert storage.storages[1].touch_field == ('status', 'my-dummy')


def test_smart_storage_with_prefix():
    storage = SmartProgressStorage(prefix='my-operator.my-company.com')
    assert isinstance(storage.storages[0], AnnotationsProgressStorage)
    assert isinstance(storage.storages[1], StatusProgressStorage)
    assert storage.storages[0].prefix == 'my-operator.my-company.com'
    assert storage.storages[1].field == ('status', 'kopf', 'progress')
    assert storage.storages[1].touch_field == ('status', 'kopf', 'dummy')


#
# Common behaviour.
#


@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_fetching_from_empty_body_returns_none(
        cls: type[ProgressStorage]):
    storage = cls()
    body = Body({})
    data = storage.fetch(body=body, key=HandlerId('id1'))
    assert data is None


@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_purging_already_empty_body_does_nothing(
        cls: type[ProgressStorage]):
    storage = cls()
    patch = Patch()
    body = Body({})
    storage.purge(body=body, patch=patch, key=HandlerId('id1'))
    assert not patch


#
# Annotations-populating.
#


@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_fetching_from_annotations_storage(cls):
    storage = cls(prefix='my-operator.example.com', verbose=True)
    body = Body({'metadata': {'annotations': {
        'my-operator.example.com/id1': CONTENT_JSON_1,
    }}})
    content = storage.fetch(body=body, key=HandlerId('id1'))

    assert content == CONTENT_DATA_1


@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_storing_to_annotations_storage_populates_keys(cls):
    storage = cls(prefix='my-operator.example.com', verbose=True)
    patch = Patch()
    body = Body({})
    storage.store(body=body, patch=patch, key=HandlerId('id1'), record=CONTENT_DATA_1)

    assert patch
    assert patch['metadata']['annotations']['my-operator.example.com/id1'] == CONTENT_JSON_1


@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_storing_to_annotations_storage_appends_keys(cls):
    storage = cls(prefix='my-operator.example.com', verbose=True)
    patch = Patch()
    body = Body({})
    storage.store(body=body, patch=patch, key=HandlerId('id1'), record=CONTENT_DATA_1)
    storage.store(body=body, patch=patch, key=HandlerId('id2'), record=CONTENT_DATA_2)

    assert patch
    assert patch['metadata']['annotations']['my-operator.example.com/id1'] == CONTENT_JSON_1
    assert patch['metadata']['annotations']['my-operator.example.com/id2'] == CONTENT_JSON_2


@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_storing_to_annotations_storage_overwrites_old_content(cls):
    storage = cls(prefix='my-operator.example.com', verbose=True)
    patch = Patch()
    body = Body({})
    storage.store(body=body, patch=patch, key=HandlerId('id1'), record=CONTENT_DATA_1)
    storage.store(body=body, patch=patch, key=HandlerId('id1'), record=CONTENT_DATA_2)

    assert patch
    assert patch['metadata']['annotations']['my-operator.example.com/id1'] == CONTENT_JSON_2


@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_storing_to_annotations_storage_cleans_content(cls):
    storage = cls(prefix='my-operator.example.com')  # no verbose=
    patch = Patch()
    body = Body({})
    content = ProgressRecord(
        started=None,
        stopped=None,
        delayed=None,
        retries=None,
        success=None,
        failure=None,
        message=None,
        subrefs=None,
    )
    storage.store(body=body, patch=patch, key=HandlerId('id1'), record=content)

    assert patch
    assert patch['metadata']['annotations']['my-operator.example.com/id1'] == json.dumps({})


@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_purging_of_annotations_storage_nullifies_content(cls):
    storage = cls(prefix='my-operator.example.com', verbose=True)
    patch = Patch()
    body = Body({'metadata': {'annotations': {
        'my-operator.example.com/id1': CONTENT_JSON_1,
    }}})
    storage.purge(body=body, patch=patch, key=HandlerId('id1'))

    assert patch
    assert patch['metadata']['annotations']['my-operator.example.com/id1'] is None


@pytest.mark.parametrize('body_data', [
    pytest.param({}, id='without-data'),
    pytest.param({'metadata': {'annotations': {'my-operator.example.com/my-dummy': 'something'}}}, id='with-data'),
])
@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_touching_via_annotations_storage_with_payload(cls, body_data):
    storage = cls(prefix='my-operator.example.com', touch_key='my-dummy')
    patch = Patch()
    body = Body(body_data)
    storage.touch(body=body, patch=patch, value='hello')

    assert patch
    assert patch['metadata']['annotations']['my-operator.example.com/my-dummy'] == 'hello'


@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_touching_via_annotations_storage_with_none_when_absent(cls):
    storage = cls(prefix='my-operator.example.com', touch_key='my-dummy')
    patch = Patch()
    body = Body({})
    storage.touch(body=body, patch=patch, value=None)

    assert not patch


@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_touching_via_annotations_storage_with_none_when_present(cls):
    storage = cls(prefix='my-operator.example.com', touch_key='my-dummy')
    patch = Patch()
    body = Body({'metadata': {'annotations': {'my-operator.example.com/my-dummy': 'something'}}})
    storage.touch(body=body, patch=patch, value=None)

    assert patch
    assert patch['metadata']['annotations']['my-operator.example.com/my-dummy'] is None


#
# Status-populating.
#


@pytest.mark.parametrize('cls', STATUS_POPULATING_STORAGES)
def test_fetching_from_status_storage(cls):
    storage = cls(field='status.my-operator')
    body = Body({'status': {'my-operator': {'id1': CONTENT_DATA_1, 'id2': CONTENT_DATA_2}}})
    content = storage.fetch(body=body, key=HandlerId('id1'))

    assert content == CONTENT_DATA_1


@pytest.mark.parametrize('cls', STATUS_POPULATING_STORAGES)
def test_storing_to_status_storage_populates_keys(cls):
    storage = cls(field='status.my-operator')
    patch = Patch()
    body = Body({})
    storage.store(body=body, patch=patch, key=HandlerId('id1'), record=CONTENT_DATA_1)

    assert patch
    assert patch['status']['my-operator']['id1'] == CONTENT_DATA_1


@pytest.mark.parametrize('cls', STATUS_POPULATING_STORAGES)
def test_storing_to_status_storage_appends_keys(cls):
    storage = cls(field='status.my-operator')
    patch = Patch()
    body = Body({})
    storage.store(body=body, patch=patch, key=HandlerId('id1'), record=CONTENT_DATA_1)
    storage.store(body=body, patch=patch, key=HandlerId('id2'), record=CONTENT_DATA_1)

    assert patch
    assert patch['status']['my-operator']['id1'] == CONTENT_DATA_1
    assert patch['status']['my-operator']['id2'] == CONTENT_DATA_1


@pytest.mark.parametrize('cls', STATUS_POPULATING_STORAGES)
def test_storing_to_status_storage_overwrites_old_content(cls):
    storage = cls(field='status.my-operator')
    patch = Patch()
    body = Body({})
    storage.store(body=body, patch=patch, key=HandlerId('id1'), record=CONTENT_DATA_1)
    storage.store(body=body, patch=patch, key=HandlerId('id1'), record=CONTENT_DATA_2)

    assert patch
    assert patch['status']['my-operator']['id1'] == CONTENT_DATA_2


@pytest.mark.parametrize('cls', STATUS_POPULATING_STORAGES)
def test_purging_of_status_storage_nullifies_content(cls):
    storage = cls(field='status.my-operator')
    patch = Patch()
    body = Body({'status': {'my-operator': {'id1': CONTENT_DATA_1}}})
    storage.purge(body=body, patch=patch, key=HandlerId('id1'))

    assert patch
    assert patch['status']['my-operator']['id1'] is None


@pytest.mark.parametrize('body_data', [
    pytest.param({}, id='without-data'),
    pytest.param({'status': {'my-dummy': 'something'}}, id='with-data'),
])
@pytest.mark.parametrize('cls', STATUS_POPULATING_STORAGES)
def test_touching_via_status_storage_with_payload(cls, body_data):
    storage = cls(field='status.my-operator', touch_field='status.my-dummy')
    patch = Patch()
    body = Body(body_data)
    storage.touch(body=body, patch=patch, value='hello')

    assert patch
    assert patch['status']['my-dummy'] == 'hello'


@pytest.mark.parametrize('cls', STATUS_POPULATING_STORAGES)
def test_touching_via_status_storage_with_none_when_absent(cls):
    storage = cls(touch_field='status.my-dummy')
    patch = Patch()
    body = Body({})
    storage.touch(body=body, patch=patch, value=None)

    assert not patch


@pytest.mark.parametrize('cls', STATUS_POPULATING_STORAGES)
def test_touching_via_status_storage_with_none_when_present(cls):
    storage = cls(touch_field='status.my-dummy')
    patch = Patch()
    body = Body({'status': {'my-dummy': 'something'}})
    storage.touch(body=body, patch=patch, value=None)

    assert patch
    assert patch['status']['my-dummy'] is None



import asyncio

import kopf
from kopf._core.actions.execution import ErrorsMode, PermanentError, TemporaryError


async def test_daemon_stopped_on_permanent_error(
        settings, resource, dummy, assert_logs, k8s_mocked, simulate_cycle, looptime):

    @kopf.daemon(*resource, id='fn', backoff=1.23)
    async def fn(**kwargs):
        dummy.mock(**kwargs)
        raise PermanentError("boo!")

    finalizer = settings.persistence.finalizer
    event_object = {'metadata': {'finalizers': [finalizer]}}
    await simulate_cycle(event_object)
    await asyncio.sleep(123)  # give it enough opportunities to misbehave (e.g. restart)

    assert looptime == 123
    assert dummy.mock.call_count == 1

    assert_logs([
        "Daemon 'fn' failed permanently: boo!",
        "Daemon 'fn' has exited on its own",
    ], prohibited=[
        "Daemon 'fn' succeeded.",
    ])


async def test_daemon_stopped_on_arbitrary_errors_with_mode_permanent(
        settings, resource, dummy, assert_logs, k8s_mocked, simulate_cycle, looptime):

    @kopf.daemon(*resource, id='fn', errors=ErrorsMode.PERMANENT, backoff=1.23)
    async def fn(**kwargs):
        dummy.mock(**kwargs)
        raise Exception("boo!")

    finalizer = settings.persistence.finalizer
    event_object = {'metadata': {'finalizers': [finalizer]}}
    await simulate_cycle(event_object)
    await asyncio.sleep(123)  # give it enough opportunities to misbehave (e.g. restart)

    assert looptime == 123
    assert dummy.mock.call_count == 1

    assert_logs([
        "Daemon 'fn' failed with an exception and will stop now: boo!",
        "Daemon 'fn' has exited on its own",
    ], prohibited=[
        "Daemon 'fn' succeeded.",
    ])


async def test_daemon_retried_on_temporary_error(
        registry, settings, resource, dummy,
        assert_logs, k8s_mocked, simulate_cycle, looptime):
    finished = asyncio.Event()

    @kopf.daemon(*resource, id='fn', backoff=1.23)
    async def fn(retry, **kwargs):
        dummy.mock(**kwargs)
        if not retry:
            raise TemporaryError("boo!", delay=3.45)
        else:
            finished.set()

    finalizer = settings.persistence.finalizer
    event_object = {'metadata': {'finalizers': [finalizer]}}
    await simulate_cycle(event_object)
    await finished.wait()

    assert looptime == 3.45
    assert_logs([
        "Daemon 'fn' failed temporarily: boo!",
        "Daemon 'fn' succeeded.",
        "Daemon 'fn' has exited on its own",
    ])


async def test_daemon_retried_on_arbitrary_error_with_mode_temporary(
        settings, resource, dummy, assert_logs, k8s_mocked, simulate_cycle, looptime):
    finished = asyncio.Event()

    @kopf.daemon(*resource, id='fn', errors=ErrorsMode.TEMPORARY, backoff=1.23)
    async def fn(retry, **kwargs):
        dummy.mock(**kwargs)
        if not retry:
            raise Exception("boo!")
        else:
            finished.set()

    finalizer = settings.persistence.finalizer
    event_object = {'metadata': {'finalizers': [finalizer]}}
    await simulate_cycle(event_object)
    await finished.wait()

    assert looptime == 1.23
    assert_logs([
        "Daemon 'fn' failed with an exception and will try again in 1.23 seconds: boo!",
        "Daemon 'fn' succeeded.",
        "Daemon 'fn' has exited on its own",
    ])


async def test_daemon_retried_until_retries_limit(
        resource, dummy, caplog, assert_logs, k8s_mocked, simulate_cycle, looptime):
    trigger = asyncio.Condition()

    @kopf.daemon(*resource, id='fn', retries=3)
    async def fn(**kwargs):
        dummy.mock(**kwargs)
        async with trigger:
            trigger.notify_all()
        raise TemporaryError("boo!", delay=1.23)

    await simulate_cycle({})
    async with trigger:
        await trigger.wait_for(lambda: any("but will" in m for m in caplog.messages))

    assert looptime == 2.46
    assert dummy.mock.call_count == 3


async def test_daemon_retried_until_timeout(
        resource, dummy, caplog, assert_logs, k8s_mocked, simulate_cycle, looptime):
    trigger = asyncio.Condition()

    @kopf.daemon(*resource, id='fn', timeout=4)
    async def fn(**kwargs):
        dummy.mock(**kwargs)
        async with trigger:
            trigger.notify_all()
        raise TemporaryError("boo!", delay=1.23)

    await simulate_cycle({})
    async with trigger:
        await trigger.wait_for(lambda: any("but will" in m for m in caplog.messages))

    assert looptime == 3.69
    assert dummy.mock.call_count == 4
